import logging
from threading import Lock
from threading import Timer

from pimdm.utils import TYPE_CHECKING
from .wrapper import NoListenersPresent
from .mld_globals import MulticastListenerInterval, LastListenerQueryInterval

if TYPE_CHECKING:
    from .RouterState import RouterState


class GroupState(object):
    LOGGER = logging.getLogger('pim.mld.RouterState.GroupState')

    def __init__(self, router_state: 'RouterState', group_ip: str):
        #logger
        extra_dict_logger = router_state.router_state_logger.extra.copy()
        extra_dict_logger['tree'] = '(*,' + group_ip + ')'
        self.group_state_logger = logging.LoggerAdapter(GroupState.LOGGER, extra_dict_logger)

        #timers and state
        self.router_state = router_state
        self.group_ip = group_ip
        self.state = NoListenersPresent
        self.timer = None
        self.retransmit_timer = None
        # lock
        self.lock = Lock()

        # KernelEntry's instances to notify change of igmp state
        self.multicast_interface_state = []
        self.multicast_interface_state_lock = Lock()

    def print_state(self):
        return self.state.print_state()

    ###########################################
    # Set state
    ###########################################
    def set_state(self, state):
        self.state = state
        self.group_state_logger.debug("change membership state to: " + state.print_state())

    ###########################################
    # Set timers
    ###########################################
    def set_timer(self, alternative: bool=False, max_response_time: int=None):
        self.clear_timer()
        if not alternative:
            time = MulticastListenerInterval
        else:
            time = self.router_state.interface_state.get_group_membership_time(max_response_time)

        timer = Timer(time, self.group_membership_timeout)
        timer.start()
        self.timer = timer

    def clear_timer(self):
        if self.timer is not None:
            self.timer.cancel()

    def set_retransmit_timer(self):
        self.clear_retransmit_timer()
        retransmit_timer = Timer(LastListenerQueryInterval, self.retransmit_timeout)
        retransmit_timer.start()
        self.retransmit_timer = retransmit_timer

    def clear_retransmit_timer(self):
        if self.retransmit_timer is not None:
            self.retransmit_timer.cancel()


    ###########################################
    # Get group state from specific interface state
    ###########################################
    def get_interface_group_state(self):
        return self.state.get_state(self.router_state)

    ###########################################
    # Timer timeout
    ###########################################
    def group_membership_timeout(self):
        with self.lock:
            self.get_interface_group_state().group_membership_timeout(self)

    def retransmit_timeout(self):
        with self.lock:
            self.get_interface_group_state().retransmit_timeout(self)

    ###########################################
    # Receive Packets
    ###########################################
    def receive_report(self):
        with self.lock:
            self.get_interface_group_state().receive_report(self)

    def receive_done(self):
        with self.lock:
            self.get_interface_group_state().receive_done(self)

    def receive_group_specific_query(self, max_response_time: int):
        with self.lock:
            self.get_interface_group_state().receive_group_specific_query(self, max_response_time)

    ###########################################
    # Notify Routing
    ###########################################
    def notify_routing_add(self):
        with self.multicast_interface_state_lock:
            print("notify+", self.multicast_interface_state)
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=True)

    def notify_routing_remove(self):
        with self.multicast_interface_state_lock:
            print("notify-", self.multicast_interface_state)
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=False)

    def add_multicast_routing_entry(self, kernel_entry):
        with self.multicast_interface_state_lock:
            self.multicast_interface_state.append(kernel_entry)
            return self.has_members()

    def remove_multicast_routing_entry(self, kernel_entry):
        with self.multicast_interface_state_lock:
            self.multicast_interface_state.remove(kernel_entry)

    def has_members(self):
        return self.state is not NoListenersPresent

    def remove(self):
        with self.multicast_interface_state_lock:
            self.clear_retransmit_timer()
            self.clear_timer()
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=False)
            del self.multicast_interface_state[:]



from pimdm.utils import TYPE_CHECKING
from ..wrapper import ListenersPresent
if TYPE_CHECKING:
    from ..GroupState import GroupState


def receive_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier NoListenersPresent: receive_report')
    group_state.set_timer()
    group_state.set_state(ListenersPresent)

    # NOTIFY ROUTING + !!!!
    group_state.notify_routing_add()


def receive_done(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier NoListenersPresent: receive_done')
    # do nothing
    return


def receive_group_specific_query(group_state: 'GroupState', max_response_time: int):
    group_state.group_state_logger.debug('Querier NoListenersPresent: receive_group_specific_query')
    # do nothing
    return


def group_membership_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier NoListenersPresent: group_membership_timeout')
    # do nothing
    return


def retransmit_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier NoListenersPresent: retransmit_timeout')
    # do nothing
    return




import struct
import socket
from .PacketPimEncodedGroupAddress import PacketPimEncodedGroupAddress
from .PacketPimEncodedUnicastAddress import PacketPimEncodedUnicastAddress
from pimdm.tree.globals import ASSERT_CANCEL_METRIC
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|PIM Ver| Type  |   Reserved    |           Checksum            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Multicast Group Address (Encoded Group Format)        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|             Source Address (Encoded Unicast Format)           |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|R|                     Metric Preference                       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                             Metric                            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketPimAssert:
    PIM_TYPE = 5

    PIM_HDR_ASSERT = "! %ss %ss LL"
    PIM_HDR_ASSERT_WITHOUT_ADDRESS = "! LL"
    PIM_HDR_ASSERT_v4 = PIM_HDR_ASSERT % (PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN)
    PIM_HDR_ASSERT_v6 = PIM_HDR_ASSERT % (PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN_IPv6, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN_IPV6)

    PIM_HDR_ASSERT_WITHOUT_ADDRESS_LEN = struct.calcsize(PIM_HDR_ASSERT_WITHOUT_ADDRESS)
    PIM_HDR_ASSERT_v4_LEN = struct.calcsize(PIM_HDR_ASSERT_v4)
    PIM_HDR_ASSERT_v6_LEN = struct.calcsize(PIM_HDR_ASSERT_v6)

    def __init__(self, multicast_group_address: str or bytes, source_address: str or bytes, metric_preference: int or float, metric: int or float):
        if type(multicast_group_address) is bytes:
            multicast_group_address = socket.inet_ntoa(multicast_group_address)
        if type(source_address) is bytes:
            source_address = socket.inet_ntoa(source_address)
        if metric_preference > 0x7FFFFFFF:
            metric_preference = 0x7FFFFFFF
        if metric > ASSERT_CANCEL_METRIC:
            metric = ASSERT_CANCEL_METRIC
        self.multicast_group_address = multicast_group_address
        self.source_address = source_address
        self.metric_preference = metric_preference
        self.metric = metric

    def bytes(self) -> bytes:
        multicast_group_address = PacketPimEncodedGroupAddress(self.multicast_group_address).bytes()
        source_address = PacketPimEncodedUnicastAddress(self.source_address).bytes()

        msg = multicast_group_address + source_address + struct.pack(PacketPimAssert.PIM_HDR_ASSERT_WITHOUT_ADDRESS,
                                                                     0x7FFFFFFF & self.metric_preference,
                                                                     self.metric)
        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        multicast_group_addr_obj = PacketPimEncodedGroupAddress.parse_bytes(data)
        multicast_group_addr_len = len(multicast_group_addr_obj)
        data = data[multicast_group_addr_len:]

        source_addr_obj = PacketPimEncodedUnicastAddress.parse_bytes(data)
        source_addr_len = len(source_addr_obj)
        data = data[source_addr_len:]

        (metric_preference, metric) = struct.unpack(PacketPimAssert.PIM_HDR_ASSERT_WITHOUT_ADDRESS, data[:PacketPimAssert.PIM_HDR_ASSERT_WITHOUT_ADDRESS_LEN])
        pim_payload = PacketPimAssert(multicast_group_addr_obj.group_address, source_addr_obj.unicast_address, 0x7FFFFFFF & metric_preference, metric)

        return pim_payload



import struct
import socket
from .PacketPayload import PacketPayload

"""
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|     Type      |     Code      |          Checksum             |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|     Maximum Response Delay    |          Reserved             |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
+                                                               +
|                                                               |
+                       Multicast Address                       +
|                                                               |
+                                                               +
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
class PacketMLDHeader(PacketPayload):
    MLD_TYPE = 58

    MLD_HDR = "! BB H H H 16s"
    MLD_HDR_LEN = struct.calcsize(MLD_HDR)

    MULTICAST_LISTENER_QUERY_TYPE = 130
    MULTICAST_LISTENER_REPORT_TYPE = 131
    MULTICAST_LISTENER_DONE_TYPE = 132

    def __init__(self, type: int, max_resp_delay: int, group_address: str = "::"):
        # todo check type
        self.type = type
        self.max_resp_delay = max_resp_delay
        self.group_address = group_address

    def get_mld_type(self):
        return self.type

    def bytes(self) -> bytes:
        # obter mensagem e criar checksum
        msg_without_chcksum = struct.pack(PacketMLDHeader.MLD_HDR, self.type, 0, 0, self.max_resp_delay, 0,
                                          socket.inet_pton(socket.AF_INET6, self.group_address))
        #mld_checksum = checksum(msg_without_chcksum)
        #msg = msg_without_chcksum[0:2] + struct.pack("! H", mld_checksum) + msg_without_chcksum[4:]
        # checksum handled by linux kernel
        return msg_without_chcksum

    def __len__(self):
        return len(self.bytes())


    @staticmethod
    def parse_bytes(data: bytes):
        mld_hdr = data[0:PacketMLDHeader.MLD_HDR_LEN]
        if len(mld_hdr) < PacketMLDHeader.MLD_HDR_LEN:
            raise Exception("MLD packet length is lower than expected")
        (mld_type, _, _, max_resp_delay, _, group_address) = struct.unpack(PacketMLDHeader.MLD_HDR, mld_hdr)
        # checksum is handled by linux kernel
        mld_hdr = mld_hdr[PacketMLDHeader.MLD_HDR_LEN:]
        group_address = socket.inet_ntop(socket.AF_INET6, group_address)
        pkt = PacketMLDHeader(mld_type, max_resp_delay, group_address)
        return pkt



from pimdm.packet.PacketMLDHeader import PacketMLDHeader
from pimdm.utils import TYPE_CHECKING
from ..mld_globals import LastListenerQueryInterval
from ..wrapper import ListenersPresent, NoListenersPresent
if TYPE_CHECKING:
    from ..GroupState import GroupState


def receive_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier CheckingListeners: receive_report')
    group_state.set_timer()
    group_state.clear_retransmit_timer()
    group_state.set_state(ListenersPresent)

def receive_done(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier CheckingListeners: receive_done')
    # do nothing
    return


def receive_group_specific_query(group_state: 'GroupState', max_response_time: int):
    group_state.group_state_logger.debug('Querier CheckingListeners: receive_group_specific_query')
    # do nothing
    return


def group_membership_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier CheckingListeners: group_membership_timeout')
    group_state.clear_retransmit_timer()
    group_state.set_state(NoListenersPresent)

    # NOTIFY ROUTING - !!!!
    group_state.notify_routing_remove()


def retransmit_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier CheckingListeners: retransmit_timeout')
    group_addr = group_state.group_ip
    packet = PacketMLDHeader(type=PacketMLDHeader.MULTICAST_LISTENER_QUERY_TYPE,
                             max_resp_delay=LastListenerQueryInterval*1000, group_address=group_addr)
    group_state.router_state.send(data=packet.bytes(), address=group_addr)

    group_state.set_retransmit_timer()



from pimdm.utils import TYPE_CHECKING
from ..wrapper import NoMembersPresent
from ..wrapper import MembersPresent

if TYPE_CHECKING:
    from ..GroupState import GroupState


def group_membership_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: group_membership_timeout')
    group_state.set_state(NoMembersPresent)

    # NOTIFY ROUTING - !!!!
    group_state.notify_routing_remove()


def group_membership_v1_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: group_membership_v1_timeout')
    # do nothing
    return


def retransmit_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: retransmit_timeout')
    # do nothing
    return


def receive_v1_membership_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: receive_v1_membership_report')
    receive_v2_membership_report(group_state)


def receive_v2_membership_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: receive_v2_membership_report')
    group_state.set_timer()
    group_state.set_state(MembersPresent)


def receive_leave_group(group_state: 'GroupState'):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: receive_leave_group')
    # do nothing
    return


def receive_group_specific_query(group_state: 'GroupState', max_response_time: int):
    group_state.group_state_logger.debug('NonQuerier CheckingMembership: receive_group_specific_query')
    # do nothing
    return



from abc import ABCMeta, abstractmethod


class OriginatorStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def recvDataMsgFromSource(tree):
        pass

    @staticmethod
    @abstractmethod
    def SRTexpires(tree):
        pass

    @staticmethod
    @abstractmethod
    def SATexpires(tree):
        pass

    @staticmethod
    @abstractmethod
    def SourceNotConnected(tree):
        pass


class Originator(OriginatorStateABC):
    @staticmethod
    def recvDataMsgFromSource(tree):
        tree.set_source_active_timer()

    @staticmethod
    def SRTexpires(tree):
        '''
        @type tree: Tree
        '''
        tree.originator_logger.debug('SRT expired, O -> O')
        tree.set_state_refresh_timer()
        tree.create_state_refresh_msg()

    @staticmethod
    def SATexpires(tree):
        tree.originator_logger.debug('SAT expired, O -> NO')
        tree.clear_state_refresh_timer()
        tree.set_originator_state(OriginatorState.NotOriginator)

    @staticmethod
    def SourceNotConnected(tree):
        tree.originator_logger.debug('Source no longer directly connected, O -> NO')
        tree.clear_state_refresh_timer()
        tree.clear_source_active_timer()
        tree.set_originator_state(OriginatorState.NotOriginator)

    def __str__(self):
        return 'Originator'

class NotOriginator(OriginatorStateABC):
    @staticmethod
    def recvDataMsgFromSource(tree):
        '''
        @type interface: Tree
        '''
        tree.originator_logger.debug('new DataMsg from Source, NO -> O')
        tree.set_originator_state(OriginatorState.Originator)

        tree.set_state_refresh_timer()
        tree.set_source_active_timer()

    @staticmethod
    def SRTexpires(tree):
        assert False, "SRTexpires in NO"

    @staticmethod
    def SATexpires(tree):
        assert False, "SATexpires in NO"

    @staticmethod
    def SourceNotConnected(tree):
        return

    def __str__(self):
        return 'NotOriginator'


class OriginatorState():
    NotOriginator = NotOriginator()
    Originator = Originator()



import sys
import time
import netifaces
import logging
import logging.handlers
from prettytable import PrettyTable

from pimdm import UnicastRouting
from pimdm.TestLogger import RootFilter

interfaces = {}  # interfaces with multicast routing enabled
igmp_interfaces = {}  # igmp interfaces
interfaces_v6 = {}  # pim v6 interfaces
mld_interfaces = {}  # mld interfaces
kernel = None
kernel_v6 = None
unicast_routing = None
logger = None


def add_pim_interface(interface_name, state_refresh_capable: bool = False, ipv4=True, ipv6=False):
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            add_pim_interface(interface_name, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.create_pim_interface(interface_name=interface_name, state_refresh_capable=state_refresh_capable)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.create_pim_interface(interface_name=interface_name, state_refresh_capable=state_refresh_capable)


def add_membership_interface(interface_name, ipv4=True, ipv6=False):
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            add_membership_interface(interface_name, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.create_membership_interface(interface_name=interface_name)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.create_membership_interface(interface_name=interface_name)


def remove_interface(interface_name, pim=False, membership=False, ipv4=True, ipv6=False):
    if interface_name == "*":
        for interface_name in netifaces.interfaces():
            remove_interface(interface_name, pim, membership, ipv4, ipv6)
        return

    if ipv4 and kernel is not None:
        kernel.remove_interface(interface_name, pim=pim, membership=membership)
    if ipv6 and kernel_v6 is not None:
        kernel_v6.remove_interface(interface_name, pim=pim, membership=membership)


def list_neighbors(ipv4=False, ipv6=False):
    if ipv4:
        interfaces_list = interfaces.values()
    elif ipv6:
        interfaces_list = interfaces_v6.values()
    else:
        return "Unknown IP family"

    t = PrettyTable(['Interface', 'Neighbor IP', 'Hello Hold Time', "Generation ID", "Uptime"])
    check_time = time.time()
    for interface in interfaces_list:
        for neighbor in interface.get_neighbors():
            uptime = check_time - neighbor.time_of_last_update
            uptime = 0 if (uptime < 0) else uptime

            t.add_row(
                [interface.interface_name, neighbor.ip, neighbor.hello_hold_time, neighbor.generation_id, time.strftime("%H:%M:%S", time.gmtime(uptime))])
    print(t)
    return str(t)


def list_enabled_interfaces(ipv4=False, ipv6=False):
    if ipv4:
        t = PrettyTable(['Interface', 'IP', 'PIM/IGMP Enabled', 'State Refresh Enabled', 'IGMP State'])
        family = netifaces.AF_INET
        pim_interfaces = interfaces
        membership_interfaces = igmp_interfaces
    elif ipv6:
        t = PrettyTable(['Interface', 'IP', 'PIM/MLD Enabled', 'State Refresh Enabled', 'MLD State'])
        family = netifaces.AF_INET6
        pim_interfaces = interfaces_v6
        membership_interfaces = mld_interfaces
    else:
        return "Unknown IP family"

    for interface in netifaces.interfaces():
        try:
            # TODO: fix same interface with multiple ips
            ip = netifaces.ifaddresses(interface)[family][0]['addr']
            pim_enabled = interface in pim_interfaces
            membership_enabled = interface in membership_interfaces
            enabled = str(pim_enabled) + "/" + str(membership_enabled)
            state_refresh_enabled = "-"
            if pim_enabled:
                state_refresh_enabled = pim_interfaces[interface].is_state_refresh_enabled()
            membership_state = "-"
            if membership_enabled:
                membership_state = membership_interfaces[interface].interface_state.print_state()
            t.add_row([interface, ip, enabled, state_refresh_enabled, membership_state])
        except Exception:
            continue
    print(t)
    return str(t)


def list_state(ipv4=True, ipv6=False):
    state_text = ""
    if ipv4:
        state_text = "IGMP State:\n{}\n\n\n\nMulticast Routing State:\n{}"
    elif ipv6:
        state_text = "MLD State:\n{}\n\n\n\nMulticast Routing State:\n{}"
    else:
        return state_text
    return state_text.format(list_membership_state(ipv4, ipv6), list_routing_state(ipv4, ipv6))


def list_membership_state(ipv4=True, ipv6=False):
    t = PrettyTable(['Interface', 'RouterState', 'Group Adress', 'GroupState'])
    if ipv4:
        membership_interfaces = igmp_interfaces
    elif ipv6:
        membership_interfaces = mld_interfaces
    else:
        membership_interfaces = {}

    for (interface_name, interface_obj) in list(membership_interfaces.items()):
        interface_state = interface_obj.interface_state
        state_txt = interface_state.print_state()
        print(interface_state.group_state.items())

        for (group_addr, group_state) in list(interface_state.group_state.items()):
            print(group_addr)
            group_state_txt = group_state.print_state()
            t.add_row([interface_name, state_txt, group_addr, group_state_txt])
    return str(t)


def list_routing_state(ipv4=False, ipv6=False):
    if ipv4:
        routes = kernel.routing.values()
        vif_indexes = kernel.vif_index_to_name_dic.keys()
        dict_index_to_name = kernel.vif_index_to_name_dic
    elif ipv6:
        routes = kernel_v6.routing.values()
        vif_indexes = kernel_v6.vif_index_to_name_dic.keys()
        dict_index_to_name = kernel_v6.vif_index_to_name_dic
    else:
        raise Exception("Unknown IP family")

    routing_entries = []
    for a in list(routes):
        for b in list(a.values()):
            routing_entries.append(b)

    t = PrettyTable(['SourceIP', 'GroupIP', 'Interface', 'PruneState', 'AssertState', 'LocalMembership', "Is Forwarding?"])
    for entry in routing_entries:
        ip = entry.source_ip
        group = entry.group_ip
        upstream_if_index = entry.inbound_interface_index

        for index in vif_indexes:
            interface_state = entry.interface_state[index]
            interface_name = dict_index_to_name[index]
            local_membership = type(interface_state._local_membership_state).__name__
            try:
                assert_state = type(interface_state._assert_state).__name__
                if index != upstream_if_index:
                    prune_state = type(interface_state._prune_state).__name__
                    is_forwarding = interface_state.is_forwarding()
                else:
                    prune_state = type(interface_state._graft_prune_state).__name__
                    is_forwarding = "upstream"
            except:
                prune_state = "-"
                assert_state = "-"
                is_forwarding = "-"

            t.add_row([ip, group, interface_name, prune_state, assert_state, local_membership, is_forwarding])
    return str(t)


def stop():
    remove_interface("*", pim=True, membership=True, ipv4=True, ipv6=True)
    if kernel is not None:
        kernel.exit()
    if kernel_v6 is not None:
        kernel_v6.exit()
    unicast_routing.stop()


def test(router_name, server_logger_ip):
    global logger
    socketHandler = logging.handlers.SocketHandler(server_logger_ip,
                                                   logging.handlers.DEFAULT_TCP_LOGGING_PORT)
    # don't bother with a formatter, since a socket handler sends the event as
    # an unformatted pickle
    socketHandler.addFilter(RootFilter(router_name))
    logger.addHandler(socketHandler)


def enable_ipv6_kernel():
    """
    Function to explicitly enable IPv6 Multicast Routing stack.
    This may not be enabled by default due to some old linux kernels that may not have IPv6 stack or do not have
    IPv6 multicast routing support
    """
    global kernel_v6
    from pimdm.Kernel import Kernel6
    kernel_v6 = Kernel6()

    global interfaces_v6
    global mld_interfaces
    interfaces_v6 = kernel_v6.pim_interface
    mld_interfaces = kernel_v6.membership_interface


def main():
    # logging
    global logger
    logger = logging.getLogger('pim')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    global kernel
    from pimdm.Kernel import Kernel4
    kernel = Kernel4()

    global unicast_routing
    unicast_routing = UnicastRouting.UnicastRouting()

    global interfaces
    global igmp_interfaces
    interfaces = kernel.pim_interface
    igmp_interfaces = kernel.membership_interface

    try:
        enable_ipv6_kernel()
    except:
        pass



import socket
import struct
import ipaddress
import traceback
from socket import if_nametoindex
from threading import RLock, Thread
from abc import abstractmethod, ABCMeta

from pimdm import UnicastRouting, Main
from pimdm.rwlock.RWLock import RWLockWrite

from pimdm.InterfaceMLD import InterfaceMLD
from pimdm.InterfaceIGMP import InterfaceIGMP
from pimdm.InterfacePIM import InterfacePim
from pimdm.InterfacePIM6 import InterfacePim6
from pimdm.tree.KernelEntry import KernelEntry
from pimdm.tree.KernelEntryInterface import KernelEntry4Interface, KernelEntry6Interface


class Kernel(metaclass=ABCMeta):
    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    def __init__(self, kernel_socket):
        # Kernel is running
        self.running = True

        # KEY : interface_ip, VALUE : vif_index
        self.vif_index_to_name_dic = {}  # KEY : vif_index, VALUE : interface_name
        self.vif_name_to_index_dic = {}  # KEY : interface_name, VALUE : vif_index

        # KEY : source_ip, VALUE : {group_ip: KernelEntry}
        self.routing = {}

        self.socket = kernel_socket
        self.rwlock = RWLockWrite()
        self.interface_lock = RLock()

        # Create register interface
        # todo useless in PIM-DM... useful in PIM-SM
        #self.create_virtual_interface("0.0.0.0", "pimreg", index=0, flags=Kernel.VIFF_REGISTER)

        # interfaces being monitored by this process
        self.pim_interface = {}  # name: interface_pim
        self.membership_interface = {}  # name: interface_igmp or interface_mld

        # logs
        self.interface_logger = Main.logger.getChild('KernelInterface')
        self.tree_logger = Main.logger.getChild('KernelTree')

        # receive signals from kernel with a background thread
        handler_thread = Thread(target=self.handler)
        handler_thread.daemon = True
        handler_thread.start()

    '''
    Structure to create/remove virtual interfaces
    struct vifctl {
        vifi_t	vifc_vifi;		        /* Index of VIF */
        unsigned char vifc_flags;	    /* VIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
        union {
            struct in_addr vifc_lcl_addr;     /* Local interface address */
            int            vifc_lcl_ifindex;  /* Local interface index   */
        };
        struct in_addr vifc_rmt_addr;	/* IPIP tunnel addr */
    };
    '''
    @abstractmethod
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        raise NotImplementedError

    def create_pim_interface(self, interface_name: str, state_refresh_capable:bool):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = pim_interface or membership_interface
            if pim_interface:
                # already exists
                pim_interface.set_state_refresh_capable(state_refresh_capable)
                return
            elif membership_interface:
                index = membership_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            ip_interface = None
            if interface_name not in self.pim_interface:
                pim_interface = self._create_pim_interface_object(interface_name, index, state_refresh_capable)
                self.pim_interface[interface_name] = pim_interface
                ip_interface = pim_interface.ip_interface

            if not vif_already_exists:
                self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)

    @abstractmethod
    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        raise NotImplementedError

    def create_membership_interface(self, interface_name: str):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = pim_interface or membership_interface
            if membership_interface:
                # already exists
                return
            elif pim_interface:
                index = pim_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            ip_interface = None
            if interface_name not in self.membership_interface:
                igmp_interface = self._create_membership_interface_object(interface_name, index)
                self.membership_interface[interface_name] = igmp_interface
                ip_interface = igmp_interface.ip_interface

            if not vif_already_exists:
                self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)

    @abstractmethod
    def _create_membership_interface_object(self, interface_name, index):
        raise NotImplementedError

    def remove_interface(self, interface_name, membership: bool = False, pim: bool = False):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            if (membership and not membership_interface) or (pim and not pim_interface) or (not membership and not pim):
                return
            if pim:
                pim_interface = self.pim_interface.pop(interface_name)
                pim_interface.remove()
            elif membership:
                membership_interface = self.membership_interface.pop(interface_name)
                membership_interface.remove()

            if not self.membership_interface.get(interface_name) and not self.pim_interface.get(interface_name):
                self.remove_virtual_interface(interface_name)

    @abstractmethod
    def remove_virtual_interface(self, interface_name):
        raise NotImplementedError

    #############################################
    # Manipulate multicast routing table
    #############################################
    @abstractmethod
    def set_multicast_route(self, kernel_entry: KernelEntry):
        raise NotImplementedError

    @abstractmethod
    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        raise NotImplementedError

    @abstractmethod
    def remove_multicast_route(self, kernel_entry: KernelEntry):
        raise NotImplementedError

    @abstractmethod
    def exit(self):
        raise NotImplementedError

    @abstractmethod
    def handler(self):
        raise NotImplementedError

    def get_routing_entry(self, source_group: tuple, create_if_not_existent=True):
        ip_src = source_group[0]
        ip_dst = source_group[1]
        with self.rwlock.genRlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                return self.routing[ip_src][ip_dst]

        with self.rwlock.genWlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                return self.routing[ip_src][ip_dst]
            elif create_if_not_existent:
                kernel_entry = KernelEntry(ip_src, ip_dst, self._get_kernel_entry_interface())
                if ip_src not in self.routing:
                    self.routing[ip_src] = {}

                iif = UnicastRouting.check_rpf(ip_src)
                self.set_flood_multicast_route(ip_src, ip_dst, iif)
                self.routing[ip_src][ip_dst] = kernel_entry
                return kernel_entry
            else:
                return None

    @staticmethod
    @abstractmethod
    def _get_kernel_entry_interface():
        pass

    # notify KernelEntries about changes at the unicast routing table
    def notify_unicast_changes(self, subnet):
        with self.rwlock.genWlock():
            for source_ip in list(self.routing.keys()):
                source_ip_obj = ipaddress.ip_address(source_ip)
                if source_ip_obj not in subnet:
                    continue
                for group_ip in list(self.routing[source_ip].keys()):
                    self.routing[source_ip][group_ip].network_update()


    # notify about changes at the interface (IP)
    '''
    def notify_interface_change(self, interface_name):
        with self.interface_lock:
            # check if interface was already added
            if interface_name not in self.vif_name_to_index_dic:
                return

            print("trying to change ip")
            pim_interface = self.pim_interface.get(interface_name)
            if pim_interface:
                old_ip = pim_interface.get_ip()
                pim_interface.change_interface()
                new_ip = pim_interface.get_ip()
                if old_ip != new_ip:
                    self.vif_dic[new_ip] = self.vif_dic.pop(old_ip)

            igmp_interface = self.igmp_interface.get(interface_name)
            if igmp_interface:
                igmp_interface.change_interface()
    '''

    # When interface changes number of neighbors verify if olist changes and prune/forward respectively
    def interface_change_number_of_neighbors(self):
        with self.rwlock.genRlock():
            for groups_dict in self.routing.values():
                for entry in groups_dict.values():
                    entry.change_at_number_of_neighbors()

    # When new neighbor connects try to resend last state refresh msg (if AssertWinner)
    def new_or_reset_neighbor(self, vif_index, neighbor_ip):
        with self.rwlock.genRlock():
            for groups_dict in self.routing.values():
                for entry in groups_dict.values():
                    entry.new_or_reset_neighbor(vif_index, neighbor_ip)


class Kernel4(Kernel):
    # MRT
    MRT_BASE    = 200
    MRT_INIT    = (MRT_BASE)      # /* Activate the kernel mroute code 	*/
    MRT_DONE    = (MRT_BASE + 1)  # /* Shutdown the kernel mroute		*/
    MRT_ADD_VIF = (MRT_BASE + 2)  # /* Add a virtual interface		    */
    MRT_DEL_VIF = (MRT_BASE + 3)  # /* Delete a virtual interface		*/
    MRT_ADD_MFC = (MRT_BASE + 4)  # /* Add a multicast forwarding entry	*/
    MRT_DEL_MFC = (MRT_BASE + 5)  # /* Delete a multicast forwarding entry	*/
    MRT_VERSION = (MRT_BASE + 6)  # /* Get the kernel multicast version	*/
    MRT_ASSERT  = (MRT_BASE + 7)  # /* Activate PIM assert mode		    */
    MRT_PIM     = (MRT_BASE + 8)  # /* enable PIM code			        */
    MRT_TABLE   = (MRT_BASE + 9)  # /* Specify mroute table ID		    */
    #MRT_ADD_MFC_PROXY = (MRT_BASE + 10)  # /* Add a (*,*|G) mfc entry	*/
    #MRT_DEL_MFC_PROXY = (MRT_BASE + 11)  # /* Del a (*,*|G) mfc entry	*/
    #MRT_MAX = (MRT_BASE + 11)

    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    # SIGNAL MSG TYPE
    IGMPMSG_NOCACHE = 1
    IGMPMSG_WRONGVIF = 2
    IGMPMSG_WHOLEPKT = 3  # NOT USED ON PIM-DM

    # Interface flags
    VIFF_TUNNEL      = 0x1  # IPIP tunnel
    VIFF_SRCRT       = 0x2  # NI
    VIFF_REGISTER    = 0x4  # register vif
    VIFF_USE_IFINDEX = 0x8  # use vifc_lcl_ifindex instead of vifc_lcl_addr to find an interface

    def __init__(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IGMP)

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IP, self.MRT_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IP, self.MRT_PIM, 0)
        s.setsockopt(socket.IPPROTO_IP, self.MRT_ASSERT, 1)

        super().__init__(s)

    '''
    Structure to create/remove virtual interfaces
    struct vifctl {
        vifi_t	vifc_vifi;		        /* Index of VIF */
        unsigned char vifc_flags;	    /* VIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
        union {
            struct in_addr vifc_lcl_addr;     /* Local interface address */
            int            vifc_lcl_ifindex;  /* Local interface index   */
        };
        struct in_addr vifc_rmt_addr;	/* IPIP tunnel addr */
    };
    '''
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        if type(ip_interface) is str:
            ip_interface = socket.inet_aton(ip_interface)

        struct_mrt_add_vif = struct.pack("HBBI 4s 4s", index, flags, 1, 0, ip_interface,
                                         socket.inet_aton("0.0.0.0"))
        with self.rwlock.genWlock():
            self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_VIF, struct_mrt_add_vif)
            self.vif_index_to_name_dic[index] = interface_name
            self.vif_name_to_index_dic[interface_name] = index

            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def remove_virtual_interface(self, interface_name):
        #with self.interface_lock:
        index = self.vif_name_to_index_dic.pop(interface_name, None)
        struct_vifctl = struct.pack("HBBI 4s 4s", index, 0, 0, 0, socket.inet_aton("0.0.0.0"), socket.inet_aton("0.0.0.0"))

        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DEL_VIF, struct_vifctl)

        del self.vif_name_to_index_dic[self.vif_index_to_name_dic[index]]
        interface_name = self.vif_index_to_name_dic.pop(index)

        # change MFC's to not forward traffic by this interface (set OIL to 0 for this interface)
        with self.rwlock.genWlock():
            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.remove_interface(index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, index)



    '''
    /* Cache manipulation structures for mrouted and PIMd */
    struct mfcctl {
        struct in_addr mfcc_origin;		    /* Origin of mcast	*/
        struct in_addr mfcc_mcastgrp;		/* Group in question	*/
        vifi_t mfcc_parent; 			    /* Where it arrived	*/
        unsigned char mfcc_ttls[MAXVIFS];	/* Where it is going	*/
        unsigned int mfcc_pkt_cnt;		    /* pkt count for src-grp */
        unsigned int mfcc_byte_cnt;
        unsigned int mfcc_wrong_if;
        int mfcc_expire;
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != Kernel.MAXVIFS:
            raise Exception

        #outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*4

        #outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        #struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, kernel_entry.inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_MFC, struct_mfcctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        source_ip = socket.inet_aton(source_ip)
        group_ip = socket.inet_aton(group_ip)

        outbound_interfaces = [1]*self.MAXVIFS
        outbound_interfaces[inbound_interface_index] = 0

        #outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*3 + [20]

        #outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        #struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_MFC, struct_mfcctl)

    def remove_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)
        outbound_interfaces_and_other_parameters = [0] + [0]*Kernel.MAXVIFS + [0]*4

        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DEL_MFC, struct_mfcctl)
        self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
        if len(self.routing[kernel_entry.source_ip]) == 0:
            self.routing.pop(kernel_entry.source_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DONE, 1)
        self.socket.close()


    '''
    /* This is the format the mroute daemon expects to see IGMP control
    * data. Magically happens to be like an IP packet as per the original
    */
    struct igmpmsg {
        __u32 unused1,unused2;
        unsigned char im_msgtype;		/* What is this */
        unsigned char im_mbz;			/* Must be zero */
        unsigned char im_vif;			/* Interface (this ought to be a vifi_t!) */
        unsigned char unused3;
        struct in_addr im_src,im_dst;
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(20)
                (_, _, im_msgtype, im_mbz, im_vif, _, im_src, im_dst) = struct.unpack("II B B B B 4s 4s", msg[:20])
                print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im_mbz != 0:
                    continue

                print(im_msgtype)
                print(im_mbz)
                print(im_vif)
                print(socket.inet_ntoa(im_src))
                print(socket.inet_ntoa(im_dst))
                #print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntoa(im_src)
                ip_dst = socket.inet_ntoa(im_dst)

                if im_msgtype == self.IGMPMSG_NOCACHE:
                    print("IGMP NO CACHE")
                    self.igmpmsg_nocache_handler(ip_src, ip_dst, im_vif)
                elif im_msgtype == self.IGMPMSG_WRONGVIF:
                    print("WRONG VIF HANDLER")
                    self.igmpmsg_wrongvif_handler(ip_src, ip_dst, im_vif)
                #elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def igmpmsg_nocache_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    # receive multicast (S,G) packet in a outbound_interface
    def igmpmsg_wrongvif_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)


    ''' useless in PIM-DM... useful in PIM-SM
    def igmpmsg_wholepacket_handler(self, ip_src, ip_dst):
        #kernel_entry = self.routing[(ip_src, ip_dst)]
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg()
        #kernel_entry.recv_data_msg(iif)
    '''

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry4Interface

    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        return InterfacePim(interface_name, index, state_refresh_capable)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceIGMP(interface_name, index)


class Kernel6(Kernel):
    # MRT6
    MRT6_BASE = 200
    MRT6_INIT = (MRT6_BASE)                 # /* Activate the kernel mroute code 	*/
    MRT6_DONE = (MRT6_BASE + 1)             # /* Shutdown the kernel mroute		*/
    MRT6_ADD_MIF = (MRT6_BASE + 2)          # /* Add a virtual interface		*/
    MRT6_DEL_MIF = (MRT6_BASE + 3)          # /* Delete a virtual interface		*/
    MRT6_ADD_MFC = (MRT6_BASE + 4)          # /* Add a multicast forwarding entry	*/
    MRT6_DEL_MFC = (MRT6_BASE + 5)          # /* Delete a multicast forwarding entry	*/
    MRT6_VERSION = (MRT6_BASE + 6)          # /* Get the kernel multicast version	*/
    MRT6_ASSERT = (MRT6_BASE + 7)           # /* Activate PIM assert mode		*/
    MRT6_PIM = (MRT6_BASE + 8)              # /* enable PIM code			*/
    MRT6_TABLE = (MRT6_BASE + 9)            # /* Specify mroute table ID		*/
    MRT6_ADD_MFC_PROXY = (MRT6_BASE + 10)   # /* Add a (*,*|G) mfc entry	*/
    MRT6_DEL_MFC_PROXY = (MRT6_BASE + 11)   # /* Del a (*,*|G) mfc entry	*/
    MRT6_MAX = (MRT6_BASE + 11)

    # define SIOCGETMIFCNT_IN6	SIOCPROTOPRIVATE	/* IP protocol privates */
    # define SIOCGETSGCNT_IN6	(SIOCPROTOPRIVATE+1)
    # define SIOCGETRPF	(SIOCPROTOPRIVATE+2)

    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    # SIGNAL MSG TYPE
    MRT6MSG_NOCACHE = 1
    MRT6MSG_WRONGMIF = 2
    MRT6MSG_WHOLEPKT = 3  # /* used for use level encap */

    # Interface flags
    MIFF_REGISTER = 0x1  # /* register vif	*/

    def __init__(self):
        s = socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_ICMPV6)

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_PIM, 0)
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ASSERT, 1)

        super().__init__(s)

    '''
    Structure to create/remove multicast interfaces
    struct mif6ctl {
        mifi_t	mif6c_mifi;		        /* Index of MIF */
        unsigned char mif6c_flags;	    /* MIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        __u16	 mif6c_pifi;		    /* the index of the physical IF */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
    };
    '''
    def create_virtual_interface(self, ip_interface, interface_name: str, index, flags=0x0):
        physical_if_index = if_nametoindex(interface_name)
        struct_mrt_add_vif = struct.pack("HBBHI", index, flags, 1, physical_if_index, 0)
        with self.rwlock.genWlock():
            self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MIF, struct_mrt_add_vif)
            self.vif_index_to_name_dic[index] = interface_name
            self.vif_name_to_index_dic[interface_name] = index

            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def remove_virtual_interface(self, interface_name):
        # with self.interface_lock:
        mif_index = self.vif_name_to_index_dic.pop(interface_name, None)
        interface_name = self.vif_index_to_name_dic.pop(mif_index)

        physical_if_index = if_nametoindex(interface_name)
        struct_vifctl = struct.pack("HBBHI", mif_index, 0, 0, physical_if_index, 0)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DEL_MIF, struct_vifctl)

        # alterar MFC's para colocar a 0 esta interface
        with self.rwlock.genWlock():
            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.remove_interface(mif_index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, mif_index)

    '''
    /* Cache manipulation structures for mrouted and PIMd */
    typedef	__u32 if_mask;
    typedef struct if_set {
        if_mask ifs_bits[__KERNEL_DIV_ROUND_UP(IF_SETSIZE, NIFBITS)];
    } if_set;

    struct mf6cctl {
        struct sockaddr_in6 mf6cc_origin;		/* Origin of mcast	*/
        struct sockaddr_in6 mf6cc_mcastgrp;		/* Group in question	*/
        mifi_t	mf6cc_parent;			        /* Where it arrived	*/
        struct if_set mf6cc_ifset;		        /* Where it is going */
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != 8:
            raise Exception

        # outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        # outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*4
        outgoing_interface_list = outbound_interfaces

        # outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     kernel_entry.inbound_interface_index,
                                     *outgoing_interface_list)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MFC, struct_mf6cctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        source_ip = socket.inet_pton(socket.AF_INET6, source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = [255] * 8
        outbound_interfaces[inbound_interface_index // 32] = 0xFFFFFFFF & ~(1 << (inbound_interface_index % 32))

        # outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces_and_other_parameters)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     inbound_interface_index, *outbound_interfaces)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MFC, struct_mf6cctl)

    def remove_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)
        outbound_interfaces = [0] * 8

        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, *outbound_interfaces_and_other_parameters)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group, 0,
                                     *outbound_interfaces)

        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DEL_MFC, struct_mf6cctl)
        self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
        if len(self.routing[kernel_entry.source_ip]) == 0:
            self.routing.pop(kernel_entry.source_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DONE, 1)
        self.socket.close()

    '''
    /*
     * Structure used to communicate from kernel to multicast router.
     * We'll overlay the structure onto an MLD header (not an IPv6 heder like igmpmsg{}
     * used for IPv4 implementation). This is because this structure will be passed via an
     * IPv6 raw socket, on which an application will only receiver the payload i.e the data after
     * the IPv6 header and all the extension headers. (See section 3 of RFC 3542)
     */

    struct mrt6msg {
        __u8		im6_mbz;		/* must be zero		   */
        __u8		im6_msgtype;	/* what type of message    */
        __u16		im6_mif;		/* mif rec'd on		   */
        __u32		im6_pad;		/* padding for 64 bit arch */
        struct in6_addr	im6_src, im6_dst;
    };

    /* ip6mr netlink cache report attributes */
    enum {
        IP6MRA_CREPORT_UNSPEC,
        IP6MRA_CREPORT_MSGTYPE,
        IP6MRA_CREPORT_MIF_ID,
        IP6MRA_CREPORT_SRC_ADDR,
        IP6MRA_CREPORT_DST_ADDR,
        IP6MRA_CREPORT_PKT,
        __IP6MRA_CREPORT_MAX
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(500)
                if len(msg) < 40:
                    continue
                (im6_mbz, im6_msgtype, im6_mif, _, im6_src, im6_dst) = struct.unpack("B B H I 16s 16s", msg[:40])
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im6_mbz != 0:
                    continue

                print(im6_mbz)
                print(im6_msgtype)
                print(im6_mif)
                print(socket.inet_ntop(socket.AF_INET6, im6_src))
                print(socket.inet_ntop(socket.AF_INET6, im6_dst))
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntop(socket.AF_INET6, im6_src)
                ip_dst = socket.inet_ntop(socket.AF_INET6, im6_dst)

                if im6_msgtype == self.MRT6MSG_NOCACHE:
                    print("MRT6 NO CACHE")
                    self.msg_nocache_handler(ip_src, ip_dst, im6_mif)
                elif im6_msgtype == self.MRT6MSG_WRONGMIF:
                    print("WRONG MIF HANDLER")
                    self.msg_wrongvif_handler(ip_src, ip_dst, im6_mif)
                # elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def msg_nocache_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    # receive multicast (S,G) packet in a outbound_interface
    def msg_wrongvif_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    ''' useless in PIM-DM... useful in PIM-SM
    def msg_wholepacket_handler(self, ip_src, ip_dst):
        #kernel_entry = self.routing[(ip_src, ip_dst)]
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg()
        #kernel_entry.recv_data_msg(iif)
    '''

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry6Interface

    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        return InterfacePim6(interface_name, index, state_refresh_capable)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceMLD(interface_name, index)



from ipaddress import IPv4Address
from pimdm.utils import TYPE_CHECKING
from ..igmp_globals import Membership_Query, QueryResponseInterval, LastMemberQueryCount
from pimdm.packet.PacketIGMPHeader import PacketIGMPHeader
from pimdm.packet.ReceivedPacket import ReceivedPacket
from . import NoMembersPresent, MembersPresent, CheckingMembership

if TYPE_CHECKING:
    from ..RouterState import RouterState


class NonQuerier:

    @staticmethod
    def general_query_timeout(router_state: 'RouterState'):
        router_state.router_state_logger.debug('NonQuerier state: general_query_timeout')
        # do nothing
        return

    @staticmethod
    def other_querier_present_timeout(router_state: 'RouterState'):
        router_state.router_state_logger.debug('NonQuerier state: other_querier_present_timeout')
        #change state to Querier
        router_state.change_interface_state(querier=True)

        # send general query
        packet = PacketIGMPHeader(type=Membership_Query, max_resp_time=QueryResponseInterval*10)
        router_state.interface.send(packet.bytes())

        # set general query timer
        router_state.set_general_query_timer()

    @staticmethod
    def receive_query(router_state: 'RouterState', packet: ReceivedPacket):
        router_state.router_state_logger.debug('NonQuerier state: receive_query')
        source_ip = packet.ip_header.ip_src

        # if source ip of membership query not lower than the ip of the received interface => ignore
        if IPv4Address(source_ip) >= IPv4Address(router_state.interface.get_ip()):
            return

        # reset other present querier timer
        router_state.set_other_querier_present_timer()

    # TODO ver se existe uma melhor maneira de fazer isto
    @staticmethod
    def state_name():
        return "Non Querier"

    @staticmethod
    def get_group_membership_time(max_response_time: int):
        return (max_response_time/10.0) * LastMemberQueryCount

    # State
    @staticmethod
    def get_checking_membership_state():
        return CheckingMembership

    @staticmethod
    def get_members_present_state():
        return MembersPresent

    @staticmethod
    def get_no_members_present_state():
        return NoMembersPresent

    @staticmethod
    def get_version_1_members_present_state():
        return NonQuerier.get_members_present_state()



import abc


class PacketPayload(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def bytes(self) -> bytes:
        """Get packet payload in bytes format"""

    @abc.abstractmethod
    def __len__(self):
        """Get packet payload length"""

    @staticmethod
    @abc.abstractmethod
    def parse_bytes(data: bytes):
        """From bytes create a object payload"""



from pimdm.packet.PacketIGMPHeader import PacketIGMPHeader
from pimdm.utils import TYPE_CHECKING
from ..igmp_globals import Membership_Query, LastMemberQueryInterval
from ..wrapper import Version1MembersPresent, CheckingMembership, NoMembersPresent
if TYPE_CHECKING:
    from ..GroupState import GroupState


def group_membership_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: group_membership_timeout')
    group_state.set_state(NoMembersPresent)

    # NOTIFY ROUTING - !!!!
    group_state.notify_routing_remove()


def group_membership_v1_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: group_membership_v1_timeout')
    # do nothing
    return


def retransmit_timeout(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: retransmit_timeout')
    # do nothing
    return


def receive_v1_membership_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: receive_v1_membership_report')
    group_state.set_timer()
    group_state.set_v1_host_timer()
    group_state.set_state(Version1MembersPresent)


def receive_v2_membership_report(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: receive_v2_membership_report')
    group_state.set_timer()


def receive_leave_group(group_state: 'GroupState'):
    group_state.group_state_logger.debug('Querier MembersPresent: receive_leave_group')
    group_ip = group_state.group_ip

    group_state.set_timer(alternative=True)
    group_state.set_retransmit_timer()

    packet = PacketIGMPHeader(type=Membership_Query, max_resp_time=LastMemberQueryInterval*10, group_address=group_ip)
    group_state.router_state.send(data=packet.bytes(), address=group_ip)

    group_state.set_state(CheckingMembership)


def receive_group_specific_query(group_state: 'GroupState', max_response_time):
    group_state.group_state_logger.debug('Querier MembersPresent: receive_group_specific_query')
    # do nothing
    return



import socket
import struct
from ipaddress import IPv4Address
from ctypes import create_string_buffer, addressof
import netifaces
from pimdm.Interface import Interface
from pimdm.packet.ReceivedPacket import ReceivedPacket
from pimdm.igmp.igmp_globals import Version_1_Membership_Report, Version_2_Membership_Report, Leave_Group, Membership_Query
if not hasattr(socket, 'SO_BINDTODEVICE'):
    socket.SO_BINDTODEVICE = 25


class InterfaceIGMP(Interface):
    ETH_P_IP = 0x0800		# Internet Protocol packet
    SO_ATTACH_FILTER = 26

    FILTER_IGMP = [
        struct.pack('HBBI', 0x28, 0, 0, 0x0000000c),
        struct.pack('HBBI', 0x15, 0, 3, 0x00000800),
        struct.pack('HBBI', 0x30, 0, 0, 0x00000017),
        struct.pack('HBBI', 0x15, 0, 1, 0x00000002),
        struct.pack('HBBI', 0x6, 0, 0, 0x00040000),
        struct.pack('HBBI', 0x6, 0, 0, 0x00000000),
    ]

    def __init__(self, interface_name: str, vif_index: int):
        # SEND SOCKET
        snd_s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IGMP)

        # bind to interface
        snd_s.setsockopt(socket.SOL_SOCKET, socket.SO_BINDTODEVICE, str(interface_name + "\0").encode('utf-8'))

        # RECEIVE SOCKET
        rcv_s = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(InterfaceIGMP.ETH_P_IP))

        # receive only IGMP packets by setting a BPF filter
        bpf_filter = b''.join(InterfaceIGMP.FILTER_IGMP)
        b = create_string_buffer(bpf_filter)
        mem_addr_of_filters = addressof(b)
        fprog = struct.pack('HL', len(InterfaceIGMP.FILTER_IGMP), mem_addr_of_filters)
        rcv_s.setsockopt(socket.SOL_SOCKET, InterfaceIGMP.SO_ATTACH_FILTER, fprog)

        # bind to interface
        rcv_s.bind((interface_name, 0x0800))
        super().__init__(interface_name=interface_name, recv_socket=rcv_s, send_socket=snd_s, vif_index=vif_index)
        self.interface_enabled = True
        from pimdm.igmp.RouterState import RouterState
        self.interface_state = RouterState(self)
        super()._enable()

    def get_ip(self):
        return netifaces.ifaddresses(self.interface_name)[netifaces.AF_INET][0]['addr']

    @property
    def ip_interface(self):
        """
        Get IP of this interface
        """
        return self.get_ip()

    def send(self, data: bytes, address: str="224.0.0.1"):
        super().send(data, address)

    def _receive(self, raw_bytes, ancdata, src_addr):
        if raw_bytes:
            raw_bytes = raw_bytes[14:]
            packet = ReceivedPacket(raw_bytes, self)
            ip_src = packet.ip_header.ip_src
            if not (ip_src == "0.0.0.0" or IPv4Address(ip_src).is_multicast):
                self.PKT_FUNCTIONS.get(packet.payload.get_igmp_type(), InterfaceIGMP.receive_unknown_type)(self, packet)

    ###########################################
    # Recv packets
    ###########################################
    def receive_version_1_membership_report(self, packet):
        ip_dst = packet.ip_header.ip_dst
        igmp_group = packet.payload.group_address
        if ip_dst == igmp_group and IPv4Address(igmp_group).is_multicast:
            self.interface_state.receive_v1_membership_report(packet)

    def receive_version_2_membership_report(self, packet):
        ip_dst = packet.ip_header.ip_dst
        igmp_group = packet.payload.group_address
        if ip_dst == igmp_group and IPv4Address(igmp_group).is_multicast:
            self.interface_state.receive_v2_membership_report(packet)

    def receive_leave_group(self, packet):
        ip_dst = packet.ip_header.ip_dst
        igmp_group = packet.payload.group_address
        if ip_dst == "224.0.0.2" and IPv4Address(igmp_group).is_multicast:
            self.interface_state.receive_leave_group(packet)

    def receive_membership_query(self, packet):
        ip_dst = packet.ip_header.ip_dst
        igmp_group = packet.payload.group_address
        if (IPv4Address(igmp_group).is_multicast and ip_dst == igmp_group) or \
                (ip_dst == "224.0.0.1" and igmp_group == "0.0.0.0"):
            self.interface_state.receive_query(packet)

    def receive_unknown_type(self, packet):
        return

    PKT_FUNCTIONS = {
        Version_1_Membership_Report: receive_version_1_membership_report,
        Version_2_Membership_Report: receive_version_2_membership_report,
        Leave_Group: receive_leave_group,
        Membership_Query: receive_membership_query,
    }

    ##################
    def remove(self):
        super().remove()
        self.interface_state.remove()



from pimdm.utils import TYPE_CHECKING
if TYPE_CHECKING:
    from ..RouterState import RouterState


def get_state(router_state: 'RouterState'):
    return router_state.interface_state.get_listeners_present_state()


def print_state():
    return "ListenersPresent"

'''
def group_membership_timeout(group_state):
    get_state(group_state).group_membership_timeout(group_state)


def group_membership_v1_timeout(group_state):
    get_state(group_state).group_membership_v1_timeout(group_state)


def retransmit_timeout(group_state):
    get_state(group_state).retransmit_timeout(group_state)


def receive_v1_membership_report(group_state, packet: ReceivedPacket):
    get_state(group_state).receive_v1_membership_report(group_state, packet)


def receive_v2_membership_report(group_state, packet: ReceivedPacket):
    get_state(group_state).receive_v2_membership_report(group_state, packet)


def receive_leave_group(group_state, packet: ReceivedPacket):
    get_state(group_state).receive_leave_group(group_state, packet)


def receive_group_specific_query(group_state, packet: ReceivedPacket):
    get_state(group_state).receive_group_specific_query(group_state, packet)

'''


import struct
import socket
from .PacketPimEncodedGroupAddress import PacketPimEncodedGroupAddress
from .PacketPimEncodedSourceAddress import PacketPimEncodedSourceAddress

'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Multicast Group Address 1 (Encoded Group Format)      |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|   Number of Joined Sources    |   Number of Pruned Sources    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Joined Source Address 1 (Encoded Source Format)       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                               .                               |
|                               .                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Joined Source Address n (Encoded Source Format)       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Pruned Source Address 1 (Encoded Source Format)       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                               .                               |
|                               .                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Pruned Source Address n (Encoded Source Format)       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''


class PacketPimJoinPruneMulticastGroup:
    PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP = "! %ss HH"
    PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS = "! HH"

    PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_v4_LEN_ = struct.calcsize(
        PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP % PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN)
    PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_v6_LEN_ = struct.calcsize(
        PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP % PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN_IPv6)
    PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS_LEN = struct.calcsize(
        PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS)

    PIM_HDR_JOINED_PRUNED_SOURCE = "! %ss"
    PIM_HDR_JOINED_PRUNED_SOURCE_v4_LEN = PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN
    PIM_HDR_JOINED_PRUNED_SOURCE_v6_LEN = PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN_IPV6

    def __init__(self, multicast_group: str or bytes, joined_src_addresses: list=[], pruned_src_addresses: list=[]):
        if type(multicast_group) not in (str, bytes):
            raise Exception
        elif type(multicast_group) is bytes:
            multicast_group = socket.inet_ntoa(multicast_group)

        if type(joined_src_addresses) is not list:
            raise Exception
        if type(pruned_src_addresses) is not list:
            raise Exception

        self.multicast_group = multicast_group
        self.joined_src_addresses = joined_src_addresses
        self.pruned_src_addresses = pruned_src_addresses

    def bytes(self) -> bytes:
        multicast_group_address = PacketPimEncodedGroupAddress(self.multicast_group).bytes()
        msg = multicast_group_address + struct.pack(self.PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS,
                                                    len(self.joined_src_addresses), len(self.pruned_src_addresses))

        for joined_src_address in self.joined_src_addresses:
            joined_src_address_bytes = PacketPimEncodedSourceAddress(joined_src_address).bytes()
            msg += joined_src_address_bytes

        for pruned_src_address in self.pruned_src_addresses:
            pruned_src_address_bytes = PacketPimEncodedSourceAddress(pruned_src_address).bytes()
            msg += pruned_src_address_bytes
        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        multicast_group_addr_obj = PacketPimEncodedGroupAddress.parse_bytes(data)
        multicast_group_addr_len = len(multicast_group_addr_obj)
        data = data[multicast_group_addr_len:]

        number_join_prune_data = data[:PacketPimJoinPruneMulticastGroup.PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS_LEN]
        (number_joined_sources, number_pruned_sources) = struct.unpack(PacketPimJoinPruneMulticastGroup.PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS, number_join_prune_data)

        joined = []
        pruned = []
        data = data[PacketPimJoinPruneMulticastGroup.PIM_HDR_JOIN_PRUNE_MULTICAST_GROUP_WITHOUT_GROUP_ADDRESS_LEN:]
        for i in range(0, number_joined_sources):
            joined_obj = PacketPimEncodedSourceAddress.parse_bytes(data)
            joined_obj_len = len(joined_obj)
            data = data[joined_obj_len:]
            joined.append(joined_obj.source_address)

        for i in range(0, number_pruned_sources):
            pruned_obj = PacketPimEncodedSourceAddress.parse_bytes(data)
            pruned_obj_len = len(pruned_obj)
            data = data[pruned_obj_len:]
            pruned.append(pruned_obj.source_address)

        return PacketPimJoinPruneMulticastGroup(multicast_group_addr_obj.group_address, joined, pruned)



#MLD timers (in seconds)
RobustnessVariable = 2
QueryInterval = 125
QueryResponseInterval = 10
MulticastListenerInterval = (RobustnessVariable * QueryInterval) + (QueryResponseInterval)
OtherQuerierPresentInterval = (RobustnessVariable * QueryInterval) + 0.5 * QueryResponseInterval
StartupQueryInterval = (1/4) * QueryInterval
StartupQueryCount = RobustnessVariable
LastListenerQueryInterval = 1
LastListenerQueryCount = RobustnessVariable
UnsolicitedReportInterval = 10


# MLD msg type
MULTICAST_LISTENER_QUERY_TYPE = 130
MULTICAST_LISTENER_REPORT_TYPE = 131
MULTICAST_LISTENER_DONE_TYPE = 132


from ipaddress import IPv6Address

from pimdm.utils import TYPE_CHECKING
from ..mld_globals import LastListenerQueryInterval, LastListenerQueryCount, QueryResponseInterval

from pimdm.packet.PacketMLDHeader import PacketMLDHeader
from pimdm.packet.ReceivedPacket import ReceivedPacket
from . import CheckingListeners, ListenersPresent, NoListenersPresent

if TYPE_CHECKING:
    from ..RouterState import RouterState


class Querier:
    @staticmethod
    def general_query_timeout(router_state: 'RouterState'):
        router_state.router_state_logger.debug('Querier state: general_query_timeout')
        # send general query
        packet = PacketMLDHeader(type=PacketMLDHeader.MULTICAST_LISTENER_QUERY_TYPE,
                                 max_resp_delay=QueryResponseInterval*1000)
        router_state.interface.send(packet.bytes())

        # set general query timer
        router_state.set_general_query_timer()

    @staticmethod
    def receive_query(router_state: 'RouterState', packet: ReceivedPacket):
        router_state.router_state_logger.debug('Querier state: receive_query')
        source_ip = packet.ip_header.ip_src

        # if source ip of membership query not lower than the ip of the received interface => ignore
        if IPv6Address(source_ip) >= IPv6Address(router_state.interface.get_ip()):
            return

        # if source ip of membership query lower than the ip of the received interface => change state
        # change state of interface
        # Querier -> Non Querier
        router_state.change_interface_state(querier=False)

        # set other present querier timer
        router_state.clear_general_query_timer()
        router_state.set_other_querier_present_timer()

    @staticmethod
    def other_querier_present_timeout(router_state: 'RouterState'):
        router_state.router_state_logger.debug('Querier state: other_querier_present_timeout')
        # do nothing
        return

    # TODO ver se existe uma melhor maneira de fazer isto
    @staticmethod
    def state_name():
        return "Querier"

    @staticmethod
    def get_group_membership_time(max_response_time: int):
        return LastListenerQueryInterval * LastListenerQueryCount

    # State
    @staticmethod
    def get_checking_listeners_state():
        return CheckingListeners

    @staticmethod
    def get_listeners_present_state():
        return ListenersPresent

    @staticmethod
    def get_no_listeners_present_state():
        return NoListenersPresent



import struct
import socket
from .PacketPimEncodedUnicastAddress import PacketPimEncodedUnicastAddress
from .PacketPimJoinPruneMulticastGroup import PacketPimJoinPruneMulticastGroup
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|        Upstream Neighbor Address (Encoded Unicast Format)     |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|   Reserved    |  Num Groups   |          Hold Time            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketPimJoinPrune:
    PIM_TYPE = 3

    PIM_HDR_JOIN_PRUNE = "! %ss BBH"
    PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS = "! BBH"
    PIM_HDR_JOIN_PRUNE_v4 = PIM_HDR_JOIN_PRUNE % PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN
    PIM_HDR_JOIN_PRUNE_v6 = PIM_HDR_JOIN_PRUNE % PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN_IPV6

    PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS_LEN = struct.calcsize(PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS)
    PIM_HDR_JOIN_PRUNE_v4_LEN = struct.calcsize(PIM_HDR_JOIN_PRUNE_v4)
    PIM_HDR_JOIN_PRUNE_v6_LEN = struct.calcsize(PIM_HDR_JOIN_PRUNE_v6)

    def __init__(self, upstream_neighbor_address, hold_time):
        if type(upstream_neighbor_address) not in (str, bytes):
            raise Exception
        if type(upstream_neighbor_address) is bytes:
            upstream_neighbor_address = socket.inet_ntoa(upstream_neighbor_address)
        self.groups = []
        self.upstream_neighbor_address = upstream_neighbor_address
        self.hold_time = hold_time

    def add_multicast_group(self, group: PacketPimJoinPruneMulticastGroup):
        # TODO verificar se grupo ja esta na msg
        self.groups.append(group)

    def bytes(self) -> bytes:
        upstream_neighbor_address = PacketPimEncodedUnicastAddress(self.upstream_neighbor_address).bytes()
        msg = upstream_neighbor_address + struct.pack(PacketPimJoinPrune.PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS, 0,
                                                      len(self.groups), self.hold_time)

        for multicast_group in self.groups:
            msg += multicast_group.bytes()

        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        upstream_neighbor_addr_obj = PacketPimEncodedUnicastAddress.parse_bytes(data)
        upstream_neighbor_addr_len = len(upstream_neighbor_addr_obj)
        data = data[upstream_neighbor_addr_len:]

        (_, num_groups, hold_time) = struct.unpack(PacketPimJoinPrune.PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS,
                                                   data[:PacketPimJoinPrune.PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS_LEN])
        data = data[PacketPimJoinPrune.PIM_HDR_JOIN_PRUNE_WITHOUT_ADDRESS_LEN:]
        pim_payload = cls(upstream_neighbor_addr_obj.unicast_address, hold_time)

        for i in range(0, num_groups):
            group = PacketPimJoinPruneMulticastGroup.parse_bytes(data)
            group_len = len(group)

            pim_payload.add_multicast_group(group)
            data = data[group_len:]

        return pim_payload



import socket
import ipaddress
from threading import RLock
from socket import if_indextoname
from pyroute2 import IPDB


def get_route(ip_dst: str):
    return UnicastRouting.get_route(ip_dst)


def get_metric(ip_dst: str):
    return UnicastRouting.get_metric(ip_dst)


def check_rpf(ip_dst):
    return UnicastRouting.check_rpf(ip_dst)


def get_unicast_info(ip_dst):
    return UnicastRouting.get_unicast_info(ip_dst)


class UnicastRouting(object):
    ipdb = None
    lock = RLock()

    def __init__(self):
        #UnicastRouting.ipr = IPRoute()
        UnicastRouting.ipdb = IPDB()
        self._ipdb = UnicastRouting.ipdb
        self._ipdb.register_callback(UnicastRouting.unicast_changes, mode="post")


    # get metrics (routing preference and cost) to IP ip_dst
    @staticmethod
    def get_metric(ip_dst: str):
        (metric_administrative_distance, metric_cost, _, _, mask) = UnicastRouting.get_unicast_info(ip_dst)
        return (metric_administrative_distance, metric_cost, mask)

    # get output interface IP, used to send data to IP ip_dst
    # (root interface IP to ip_dst)
    @staticmethod
    def check_rpf(ip_dst):
        # vif index of rpf interface
        return UnicastRouting.get_unicast_info(ip_dst)[3]

    @staticmethod
    def get_route(ip_dst: str):
        ip_version = ipaddress.ip_address(ip_dst).version
        if ip_version == 4:
            family = socket.AF_INET
            full_mask = 32
        elif ip_version == 6:
            family = socket.AF_INET6
            full_mask = 128
        else:
            raise Exception("Unknown IP version")
        info = None
        with UnicastRouting.lock:
            ipdb = UnicastRouting.ipdb  # type:IPDB

            for mask_len in range(full_mask, 0, -1):
                dst_network = str(ipaddress.ip_interface(ip_dst + "/" + str(mask_len)).network)

                print(dst_network)
                if dst_network in ipdb.routes:
                    print(info)
                    if ipdb.routes[{'dst': dst_network, 'family': family}]['ipdb_scope'] != 'gc':
                        info = ipdb.routes[dst_network]
                    break
                else:
                    continue
            if not info:
                print("0.0.0.0/0 or ::/0")
                if "default" in ipdb.routes:
                    info = ipdb.routes[{'dst': 'default', 'family': family}]
            print(info)
            return info

    @staticmethod
    def get_unicast_info(ip_dst):
        metric_administrative_distance = 0xFFFFFFFF
        metric_cost = 0xFFFFFFFF
        rpf_node = ip_dst
        oif = None
        mask = 0
        with UnicastRouting.lock:
            unicast_route = UnicastRouting.get_route(ip_dst)
            if unicast_route is not None:
                oif = unicast_route.get("oif")
                next_hop = unicast_route["gateway"]
                multipaths = unicast_route["multipath"]
                #prefsrc = unicast_route.get("prefsrc")

                #rpf_node = ip_dst if (next_hop is None and prefsrc is not None) else next_hop
                rpf_node = next_hop if next_hop is not None else ip_dst
                if ipaddress.ip_address(ip_dst).version == 4:
                    highest_ip = ipaddress.ip_address("0.0.0.0")
                else:
                    highest_ip = ipaddress.ip_address("::")
                for m in multipaths:
                    if m.get("gateway", None) is None:
                        oif = m.get('oif')
                        rpf_node = ip_dst
                        break
                    elif ipaddress.ip_address(m["gateway"]) > highest_ip:
                        highest_ip = ipaddress.ip_address(m["gateway"])
                        oif = m.get('oif')
                        rpf_node = m["gateway"]

                metric_administrative_distance = unicast_route["proto"]
                metric_cost = unicast_route["priority"]
                metric_cost = metric_cost if metric_cost is not None else 0
                mask = unicast_route["dst_len"]

        interface_name = None if oif is None else if_indextoname(int(oif))
        from pimdm import Main
        if ipaddress.ip_address(ip_dst).version == 4:
            rpf_if = Main.kernel.vif_name_to_index_dic.get(interface_name)
        else:
            rpf_if = Main.kernel_v6.vif_name_to_index_dic.get(interface_name)
        return (metric_administrative_distance, metric_cost, rpf_node, rpf_if, mask)

    @staticmethod
    def unicast_changes(ipdb, msg, action):
        """
        Kernel notified about a change
        Verify the type of change and recheck all trees if necessary
        """
        print("unicast change?")
        print(action)
        UnicastRouting.lock.acquire()
        family = msg['family']
        if action == "RTM_NEWROUTE" or action == "RTM_DELROUTE":
            print(ipdb.routes)
            mask_len = msg["dst_len"]
            network_address = None
            attrs = msg["attrs"]
            print(attrs)
            for (key, value) in attrs:
                print((key, value))
                if key == "RTA_DST":
                    network_address = value
                    break
            if network_address is None and family == socket.AF_INET:
                network_address = "0.0.0.0"
            elif network_address is None and family == socket.AF_INET6:
                network_address = "::"
            print(network_address)
            print(mask_len)
            print(network_address + "/" + str(mask_len))
            subnet = ipaddress.ip_network(network_address + "/" + str(mask_len))
            print(str(subnet))
            UnicastRouting.lock.release()
            from pimdm import Main
            if family == socket.AF_INET:
                Main.kernel.notify_unicast_changes(subnet)
            elif family == socket.AF_INET6:
                Main.kernel_v6.notify_unicast_changes(subnet)
            '''
        elif action == "RTM_NEWADDR" or action == "RTM_DELADDR":
            print(action)
            print(msg)
            interface_name = None
            attrs = msg["attrs"]
            for (key, value) in attrs:
                print((key, value))
                if key == "IFA_LABEL":
                    interface_name = value
                    break
            UnicastRouting.lock.release()
            try:
                Main.kernel.notify_interface_changes(interface_name)
            except:
                import traceback
                traceback.print_exc()
                pass
            subnet = ipaddress.ip_network("0.0.0.0/0")
            Main.kernel.notify_unicast_changes(subnet)
        elif action == "RTM_NEWLINK" or action == "RTM_DELLINK":
            attrs = msg["attrs"]
            if_name = None
            operation = None
            for (key, value) in attrs:
                print((key, value))
                if key == "IFLA_IFNAME":
                    if_name = value
                elif key == "IFLA_OPERSTATE":
                    operation = value
                if if_name is not None and operation is not None:
                    break
            if if_name is not None:
                print(if_name + ": " + operation)
            UnicastRouting.lock.release()
            if operation == 'DOWN':
                Main.kernel.remove_interface(if_name, membership=True, pim=True)
            subnet = ipaddress.ip_network("0.0.0.0/0")
            Main.kernel.notify_unicast_changes(subnet)
            '''
        else:
            UnicastRouting.lock.release()

    def stop(self):
        """
        No longer monitor unicast changes....
        Invoked whenever the protocol is stopped
        """
        if self._ipdb:
            self._ipdb.release()
        if UnicastRouting.ipdb:
            UnicastRouting.ipdb = None



import socket
from .Packet import Packet
from .PacketPimHeader import PacketPimHeader
from .PacketMLDHeader import PacketMLDHeader
from .PacketIGMPHeader import PacketIGMPHeader
from .PacketIpHeader import PacketIpv4Header, PacketIpv6Header
from pimdm.utils import TYPE_CHECKING
if TYPE_CHECKING:
    from pimdm.Interface import Interface


class ReceivedPacket(Packet):
    # choose payload protocol class based on ip protocol number
    payload_protocol = {2: PacketIGMPHeader, 103: PacketPimHeader}

    def __init__(self, raw_packet: bytes, interface: 'Interface'):
        self.interface = interface

        # Parse packet and fill Packet super class
        ip_header = PacketIpv4Header.parse_bytes(raw_packet)
        protocol_number = ip_header.proto

        packet_without_ip_hdr = raw_packet[ip_header.hdr_length:]
        payload = ReceivedPacket.payload_protocol[protocol_number].parse_bytes(packet_without_ip_hdr)

        super().__init__(ip_header=ip_header, payload=payload)


class ReceivedPacket_v6(Packet):
    # choose payload protocol class based on ip protocol number
    payload_protocol_v6 = {58: PacketMLDHeader, 103: PacketPimHeader}

    def __init__(self, raw_packet: bytes, ancdata: list, src_addr: str, next_header: int, interface: 'Interface'):
        self.interface = interface

        # Parse packet and fill Packet super class
        dst_addr = "::"
        for cmsg_level, cmsg_type, cmsg_data in ancdata:
            if cmsg_level == socket.IPPROTO_IPV6 and cmsg_type == socket.IPV6_PKTINFO:
                dst_addr = socket.inet_ntop(socket.AF_INET6, cmsg_data[:16])
                break

        src_addr = src_addr[0].split("%")[0]
        ipv6_packet = PacketIpv6Header(ver=6, hop_limit=1, next_header=next_header, ip_src=src_addr, ip_dst=dst_addr)
        payload = ReceivedPacket_v6.payload_protocol_v6[next_header].parse_bytes(raw_packet)
        super().__init__(ip_header=ipv6_packet, payload=payload)



ASSERT_TIME = 180
GRAFT_RETRY_PERIOD = 3
JP_OVERRIDE_INTERVAL = 3.0
OVERRIDE_INTERVAL = 2.5
PROPAGATION_DELAY = 0.5
REFRESH_INTERVAL = 60  # State Refresh Interval
SOURCE_LIFETIME = 210
T_LIMIT = 210

HELLO_HOLD_TIME_NO_TIMEOUT = 0xFFFF
HELLO_HOLD_TIME = 160
HELLO_HOLD_TIME_TIMEOUT = 0

ASSERT_CANCEL_METRIC = 0xFFFFFFFF


from .tree_interface import TreeInterface
from .upstream_prune import UpstreamState
from threading import Timer
from pimdm.custom_timer.RemainingTimer import RemainingTimer
from .globals import *
import random
from .metric import AssertMetric
from .originator import OriginatorState, OriginatorStateABC
from pimdm.packet.PacketPimStateRefresh import PacketPimStateRefresh
import traceback
from . import data_packets_socket
import threading
import logging


class TreeInterfaceUpstream(TreeInterface):
    LOGGER = logging.getLogger('pim.KernelEntry.UpstreamInterface')

    def __init__(self, kernel_entry, interface_id):
        extra_dict_logger = kernel_entry.kernel_entry_logger.extra.copy()
        extra_dict_logger['vif'] = interface_id
        extra_dict_logger['interfacename'] = kernel_entry.get_interface_name(interface_id)
        logger = logging.LoggerAdapter(TreeInterfaceUpstream.LOGGER, extra_dict_logger)
        TreeInterface.__init__(self, kernel_entry, interface_id, logger)

        # Graft/Prune State:
        self._graft_prune_state = UpstreamState.Forward
        self._graft_retry_timer = None
        self._override_timer = None
        self._prune_limit_timer = None
        self._last_rpf = self.get_neighbor_RPF()
        self.join_prune_logger.debug('Upstream state transitions to ' + str(self._graft_prune_state))

        # Originator state
        self._originator_state = OriginatorState.NotOriginator
        self._state_refresh_timer = None
        self._source_active_timer = None
        self._prune_now_counter = 0
        self.originator_logger = logging.LoggerAdapter(TreeInterfaceUpstream.LOGGER.getChild('Originator'), extra_dict_logger)
        self.originator_logger.debug('StateRefresh state transitions to ' + str(self._originator_state))

        if self.is_S_directly_conn():
            self._graft_prune_state.sourceIsNowDirectConnect(self)
            interface = self.get_interface()
            if interface is not None and interface.is_state_refresh_enabled():
                self._originator_state.recvDataMsgFromSource(self)


        # TODO TESTE SOCKET RECV DATA PCKTS
        self.socket_is_enabled = True
        (s, g) = self.get_tree_id()
        interface_name = self.get_interface_name()
        self.socket_pkt = data_packets_socket.get_s_g_bpf_filter_code(s, g, interface_name)

        # run receive method in background
        receive_thread = threading.Thread(target=self.socket_recv)
        receive_thread.daemon = True
        receive_thread.start()

        self.logger.debug('Created UpstreamInterface')


    def socket_recv(self):
        while self.socket_is_enabled:
            try:
                self.socket_pkt.recvfrom(0)
                print("PACOTE DADOS RECEBIDO")
                self.recv_data_msg()
            except:
                traceback.print_exc()
                continue

    ##########################################
    # Set state
    ##########################################
    def set_state(self, new_state):
        with self.get_state_lock():
            if new_state != self._graft_prune_state:
                self._graft_prune_state = new_state
                self.join_prune_logger.debug('Upstream state transitions to ' + str(new_state))

                self.change_tree()
                self.evaluate_ingroup()

    def set_originator_state(self, new_state: OriginatorStateABC):
        if new_state != self._originator_state:
            self._originator_state = new_state
            self.originator_logger.debug('StateRefresh state transitions to ' + str(new_state))

    ##########################################
    # Check timers
    ##########################################
    def is_graft_retry_timer_running(self):
        return self._graft_retry_timer is not None and self._graft_retry_timer.is_alive()

    def is_override_timer_running(self):
        return self._override_timer is not None and self._override_timer.is_alive()

    def is_prune_limit_timer_running(self):
        return self._prune_limit_timer is not None and self._prune_limit_timer.is_alive()

    def remaining_prune_limit_timer(self):
        return 0 if not self._prune_limit_timer else self._prune_limit_timer.time_remaining()

    ##########################################
    # Set timers
    ##########################################
    def set_graft_retry_timer(self, time=GRAFT_RETRY_PERIOD):
        self.clear_graft_retry_timer()
        self._graft_retry_timer = Timer(time, self.graft_retry_timeout)
        self._graft_retry_timer.start()

    def clear_graft_retry_timer(self):
        if self._graft_retry_timer is not None:
            self._graft_retry_timer.cancel()

    def set_override_timer(self):
        self.clear_override_timer()
        self._override_timer = Timer(self.t_override, self.override_timeout)
        self._override_timer.start()

    def clear_override_timer(self):
        if self._override_timer is not None:
            self._override_timer.cancel()

    def set_prune_limit_timer(self, time=T_LIMIT):
        self.clear_prune_limit_timer()
        self._prune_limit_timer = RemainingTimer(time, self.prune_limit_timeout)
        self._prune_limit_timer.start()

    def clear_prune_limit_timer(self):
        if self._prune_limit_timer is not None:
            self._prune_limit_timer.cancel()

    # State Refresh timers
    def set_state_refresh_timer(self):
        self.clear_state_refresh_timer()
        self._state_refresh_timer = Timer(REFRESH_INTERVAL, self.state_refresh_timeout)
        self._state_refresh_timer.start()

    def clear_state_refresh_timer(self):
        if self._state_refresh_timer is not None:
            self._state_refresh_timer.cancel()

    def set_source_active_timer(self):
        self.clear_source_active_timer()
        self._source_active_timer = Timer(SOURCE_LIFETIME, self.source_active_timeout)
        self._source_active_timer.start()

    def clear_source_active_timer(self):
        if self._source_active_timer is not None:
            self._source_active_timer.cancel()


    ###########################################
    # Timer timeout
    ###########################################
    def graft_retry_timeout(self):
        self._graft_prune_state.GRTexpires(self)

    def override_timeout(self):
        self._graft_prune_state.OTexpires(self)

    def prune_limit_timeout(self):
        return

    # State Refresh timers
    def state_refresh_timeout(self):
        self._originator_state.SRTexpires(self)

    def source_active_timeout(self):
        self._originator_state.SATexpires(self)

    ###########################################
    # Recv packets
    ###########################################
    def recv_data_msg(self):
        if not self.is_prune_limit_timer_running() and not self.is_S_directly_conn() and self.is_olist_null():
            self._graft_prune_state.dataArrivesRPFinterface_OListNull_PLTstoped(self)
        elif self.is_S_directly_conn():
            interface = self.get_interface()
            if interface is not None and  interface.is_state_refresh_enabled():
                self._originator_state.recvDataMsgFromSource(self)


    def recv_join_msg(self, upstream_neighbor_address):
        super().recv_join_msg(upstream_neighbor_address)
        if upstream_neighbor_address == self.get_neighbor_RPF():
            self._graft_prune_state.seeJoinToRPFnbr(self)

    def recv_prune_msg(self, upstream_neighbor_address, holdtime):
        super().recv_prune_msg(upstream_neighbor_address, holdtime)
        self.set_receceived_prune_holdtime(holdtime)
        self._graft_prune_state.seePrune(self)

    def recv_graft_ack_msg(self, source_ip_of_graft_ack):
        print("GRAFT ACK!!!")
        if source_ip_of_graft_ack == self.get_neighbor_RPF():
            self._graft_prune_state.recvGraftAckFromRPFnbr(self)

    def recv_state_refresh_msg(self, received_metric: AssertMetric, prune_indicator: int):
        super().recv_state_refresh_msg(received_metric, prune_indicator)

        if self.get_neighbor_RPF() != received_metric.get_ip():
            return
        if prune_indicator == 1:
            self._graft_prune_state.stateRefreshArrivesRPFnbr_pruneIs1(self)
        elif prune_indicator == 0 and not self.is_prune_limit_timer_running():
            self._graft_prune_state.stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped(self)


    ####################################
    def create_state_refresh_msg(self):
        self._prune_now_counter+=1
        (source_ip, group_ip) = self.get_tree_id()
        ph = PacketPimStateRefresh(multicast_group_adress=group_ip,
                                   source_address=source_ip,
                                   originator_adress=self.get_ip(),
                                   metric_preference=0, metric=0, mask_len=0,
                                   ttl=256,
                                   prune_indicator_flag=0,
                                   prune_now_flag=self._prune_now_counter//3,
                                   assert_override_flag=0,
                                   interval=60)

        self._prune_now_counter %= 3
        self._kernel_entry.forward_state_refresh_msg(ph)

    ###########################################
    # Change olist
    ###########################################
    def olist_is_null(self):
        self._graft_prune_state.olistIsNowNull(self)

    def olist_is_not_null(self):
        self._graft_prune_state.olistIsNowNotNull(self)


    ###########################################
    # Changes to RPF'(s)
    ###########################################

    # caused by assert transition:
    def set_assert_state(self, new_state):
        super().set_assert_state(new_state)
        self.change_rpf(self.is_olist_null())

    # caused by unicast routing table:
    def change_on_unicast_routing(self, interface_change=False):
        self.change_rpf(self.is_olist_null(), interface_change)

        '''
        if self.is_S_directly_conn():
            self._graft_prune_state.sourceIsNowDirectConnect(self)
        else:
            self._originator_state.SourceNotConnected(self)
        '''

    def change_rpf(self, olist_is_null, interface_change=False):
        current_rpf = self.get_neighbor_RPF()
        if interface_change or self._last_rpf != current_rpf:
            self._last_rpf = current_rpf
            if olist_is_null:
                self._graft_prune_state.RPFnbrChanges_olistIsNull(self)
            else:
                self._graft_prune_state.RPFnbrChanges_olistIsNotNull(self)


    ####################################################################
    #Override
    def is_forwarding(self):
        return False

    # If new/reset neighbor is RPF neighbor => clear prune limit timer
    def new_or_reset_neighbor(self, neighbor_ip):
        if neighbor_ip == self.get_neighbor_RPF():
            self.clear_prune_limit_timer()

    #Override
    def delete(self, change_type_interface=False):
        self.socket_is_enabled = False
        self.socket_pkt.close()
        super().delete(change_type_interface)
        self.clear_graft_retry_timer()
        self.clear_assert_timer()
        self.clear_prune_limit_timer()
        self.clear_override_timer()
        self.clear_state_refresh_timer()
        self.clear_source_active_timer()

        # Clear Graft/Prune State:
        self._graft_prune_state = None

        # Clear Originator state
        self._originator_state = None

    def is_downstream(self):
        return False

    def is_originator(self):
        return self._originator_state == OriginatorState.Originator

    #-------------------------------------------------------------------------
    # Properties
    #-------------------------------------------------------------------------

    @property
    def t_override(self):
        oi = self.get_interface()._override_interval
        return random.uniform(0, oi)



from threading import Timer
import logging

from pimdm.packet.PacketIGMPHeader import PacketIGMPHeader
from pimdm.packet.ReceivedPacket import ReceivedPacket
from pimdm.utils import TYPE_CHECKING
from pimdm.rwlock.RWLock import RWLockWrite
from .querier.Querier import Querier
from .nonquerier.NonQuerier import NonQuerier
from .GroupState import GroupState
from .igmp_globals import Membership_Query, QueryResponseInterval, QueryInterval, OtherQuerierPresentInterval

if TYPE_CHECKING:
    from pimdm.InterfaceIGMP import InterfaceIGMP


class RouterState(object):
    ROUTER_STATE_LOGGER = logging.getLogger('pim.igmp.RouterState')

    def __init__(self, interface: 'InterfaceIGMP'):
        #logger
        logger_extra = dict()
        logger_extra['vif'] = interface.vif_index
        logger_extra['interfacename'] = interface.interface_name
        self.router_state_logger = logging.LoggerAdapter(RouterState.ROUTER_STATE_LOGGER, logger_extra)

        # interface of the router connected to the network
        self.interface = interface

        # state of the router (Querier/NonQuerier)
        self.interface_state = Querier

        # state of each group
        # Key: GroupIPAddress, Value: GroupState object
        self.group_state = {}
        self.group_state_lock = RWLockWrite()

        # send general query
        packet = PacketIGMPHeader(type=Membership_Query, max_resp_time=QueryResponseInterval*10)
        self.interface.send(packet.bytes())

        # set initial general query timer
        timer = Timer(QueryInterval, self.general_query_timeout)
        timer.start()
        self.general_query_timer = timer

        # present timer
        self.other_querier_present_timer = None

    # Send packet via interface
    def send(self, data: bytes, address: str):
        self.interface.send(data, address)

    ############################################
    # interface_state methods
    ############################################
    def print_state(self):
        return self.interface_state.state_name()

    def set_general_query_timer(self):
        self.clear_general_query_timer()
        general_query_timer = Timer(QueryInterval, self.general_query_timeout)
        general_query_timer.start()
        self.general_query_timer = general_query_timer

    def clear_general_query_timer(self):
        if self.general_query_timer is not None:
            self.general_query_timer.cancel()

    def set_other_querier_present_timer(self):
        self.clear_other_querier_present_timer()
        other_querier_present_timer = Timer(OtherQuerierPresentInterval, self.other_querier_present_timeout)
        other_querier_present_timer.start()
        self.other_querier_present_timer = other_querier_present_timer

    def clear_other_querier_present_timer(self):
        if self.other_querier_present_timer is not None:
            self.other_querier_present_timer.cancel()

    def general_query_timeout(self):
        self.interface_state.general_query_timeout(self)

    def other_querier_present_timeout(self):
        self.interface_state.other_querier_present_timeout(self)

    def change_interface_state(self, querier: bool):
        if querier:
            self.interface_state = Querier
            self.router_state_logger.debug('change querier state to -> Querier')
        else:
            self.interface_state = NonQuerier
            self.router_state_logger.debug('change querier state to -> NonQuerier')

    ############################################
    # group state methods
    ############################################
    def get_group_state(self, group_ip):
        with self.group_state_lock.genRlock():
            if group_ip in self.group_state:
                return self.group_state[group_ip]

        with self.group_state_lock.genWlock():
            if group_ip in self.group_state:
                group_state = self.group_state[group_ip]
            else:
                group_state = GroupState(self, group_ip)
                self.group_state[group_ip] = group_state
            return group_state

    def receive_v1_membership_report(self, packet: ReceivedPacket):
        igmp_group = packet.payload.group_address
        #if igmp_group not in self.group_state:
        #    self.group_state[igmp_group] = GroupState(self, igmp_group)

        #self.group_state[igmp_group].receive_v1_membership_report()
        self.get_group_state(igmp_group).receive_v1_membership_report()

    def receive_v2_membership_report(self, packet: ReceivedPacket):
        igmp_group = packet.payload.group_address
        #if igmp_group not in self.group_state:
        #    self.group_state[igmp_group] = GroupState(self, igmp_group)

        #self.group_state[igmp_group].receive_v2_membership_report()
        self.get_group_state(igmp_group).receive_v2_membership_report()

    def receive_leave_group(self, packet: ReceivedPacket):
        igmp_group = packet.payload.group_address
        #if igmp_group in self.group_state:
        #    self.group_state[igmp_group].receive_leave_group()
        self.get_group_state(igmp_group).receive_leave_group()

    def receive_query(self, packet: ReceivedPacket):
        self.interface_state.receive_query(self, packet)
        igmp_group = packet.payload.group_address

        # process group specific query
        if igmp_group != "0.0.0.0" and igmp_group in self.group_state:
        #if igmp_group != "0.0.0.0":
            max_response_time = packet.payload.max_resp_time
            #self.group_state[igmp_group].receive_group_specific_query(max_response_time)
            self.get_group_state(igmp_group).receive_group_specific_query(max_response_time)

    def remove(self):
        for group in self.group_state.values():
            group.remove()


import struct
from pimdm.utils import checksum
import socket
from .PacketPayload import PacketPayload
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|      Type     | Max Resp Time |           Checksum            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Group Address                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
| Resv  |S| QRV |     QQIC      |     Number of Sources (N)     |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Source Address [1]                      |
+-                                                             -+
|                       Source Address [2]                      |
+-                              .                              -+
.                               .                               .
.                               .                               .
+-                                                             -+
|                       Source Address [N]                      |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketIGMPHeader(PacketPayload):
    IGMP_TYPE = 2

    IGMP_HDR = "! BB H 4s"
    IGMP_HDR_LEN = struct.calcsize(IGMP_HDR)

    IGMP3_SRC_ADDR_HDR = "! BB H "
    IGMP3_SRC_ADDR_HDR_LEN = struct.calcsize(IGMP3_SRC_ADDR_HDR)

    IPv4_HDR = "! 4s"
    IPv4_HDR_LEN = struct.calcsize(IPv4_HDR)

    Membership_Query = 0x11
    Version_2_Membership_Report = 0x16
    Leave_Group = 0x17
    Version_1_Membership_Report = 0x12

    def __init__(self, type: int, max_resp_time: int, group_address: str="0.0.0.0"):
        # todo check type
        self.type = type
        self.max_resp_time = max_resp_time
        self.group_address = group_address

    def get_igmp_type(self):
        return self.type

    def bytes(self) -> bytes:
        # obter mensagem e criar checksum
        msg_without_chcksum = struct.pack(PacketIGMPHeader.IGMP_HDR, self.type, self.max_resp_time, 0,
                                          socket.inet_aton(self.group_address))
        igmp_checksum = checksum(msg_without_chcksum)
        msg = msg_without_chcksum[0:2] + struct.pack("! H", igmp_checksum) + msg_without_chcksum[4:]
        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        #print("parseIGMPHdr: ", data)

        igmp_hdr = data[0:PacketIGMPHeader.IGMP_HDR_LEN]
        (type, max_resp_time, rcv_checksum, group_address) = struct.unpack(PacketIGMPHeader.IGMP_HDR, igmp_hdr)

        #print(type, max_resp_time, rcv_checksum, group_address)


        msg_to_checksum = data[0:2] + b'\x00\x00' + data[4:]
        #print("checksum calculated: " + str(checksum(msg_to_checksum)))
        if checksum(msg_to_checksum) != rcv_checksum:
            #print("wrong checksum")
            raise Exception("wrong checksum")

        igmp_hdr = igmp_hdr[PacketIGMPHeader.IGMP_HDR_LEN:]
        group_address = socket.inet_ntoa(group_address)
        pkt = PacketIGMPHeader(type, max_resp_time, group_address)
        return pkt


from abc import ABCMeta, abstractmethod

from . import globals as pim_globals
from pimdm.utils import TYPE_CHECKING
if TYPE_CHECKING:
    from .tree_if_downstream import TreeInterfaceDownstream

class DownstreamStateABS(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def receivedPrune(interface: "TreeInterfaceDownstream", holdtime):
        """
        Receive Prune(S,G)

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedJoin(interface: "TreeInterfaceDownstream"):
        """
        Receive Join(S,G)

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedGraft(interface: "TreeInterfaceDownstream", source_ip):
        """
        Receive Graft(S,G)

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def PPTexpires(interface: "TreeInterfaceDownstream"):
        """
        PPT(S,G) Expires

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def PTexpires(interface: "TreeInterfaceDownstream"):
        """
        PT(S,G) Expires

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def is_now_RPF_Interface(interface: "TreeInterfaceDownstream"):
        """
        RPF_Interface(S) becomes I

        @type interface: Downstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def send_state_refresh(interface: "TreeInterfaceDownstream"):
        """
        Send State Refresh(S,G) out I

        @type interface: Downstream
        """
        raise NotImplementedError()

    def __str__(self):
        return "Downstream." + self.__class__.__name__


class NoInfo(DownstreamStateABS):
    '''
    NoInfo(NI)
    The interface has no (S,G) Prune state, and neither the Prune
    timer (PT(S,G,I)) nor the PrunePending timer ((PPT(S,G,I)) is
    running.
    '''

    @staticmethod
    def receivedPrune(interface: "TreeInterfaceDownstream", holdtime):
        """
        Receive Prune(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug("receivedPrune, NI -> PP")
        interface.set_prune_state(DownstreamState.PrunePending)

        time = 0
        if len(interface.get_interface().neighbors) > 1:
            time = pim_globals.JP_OVERRIDE_INTERVAL

        interface.set_prune_pending_timer(time)

    @staticmethod
    def receivedJoin(interface: "TreeInterfaceDownstream"):
        """
        Receive Join(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        # Do nothing
        interface.join_prune_logger.debug("receivedJoin, NI -> NI")

    @staticmethod
    def receivedGraft(interface: "TreeInterfaceDownstream", source_ip):
        """
        Receive Graft(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedGraft, NI -> NI')
        interface.send_graft_ack(source_ip)

    @staticmethod
    def PPTexpires(interface: "TreeInterfaceDownstream"):
        """
        PPT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        #assert False, "PPTexpires in state NI"
        return

    @staticmethod
    def PTexpires(interface: "TreeInterfaceDownstream"):
        """
        PT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        #assert False, "PTexpires in state NI"
        return

    @staticmethod
    def is_now_RPF_Interface(interface: "TreeInterfaceDownstream"):
        """
        RPF_Interface(S) becomes I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        # Do nothing
        return

    @staticmethod
    def send_state_refresh(interface: "TreeInterfaceDownstream"):
        """
        Send State Refresh(S,G) out I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        # Do nothing
        return

    def __str__(self):
        return "NoInfo"



class PrunePending(DownstreamStateABS):
    '''
    PrunePending(PP)
    The router has received a Prune(S,G) on this interface from a
    downstream neighbor and is waiting to see whether the prune will
    be overridden by another downstream router. For forwarding
    purposes, the PrunePending state functions exactly like the
    NoInfo state.
    '''

    @staticmethod
    def receivedPrune(interface: "TreeInterfaceDownstream", holdtime):
        """
        Receive Prune(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedPrune, PP -> PP')


    @staticmethod
    def receivedJoin(interface: "TreeInterfaceDownstream"):
        """
        Receive Join(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedJoin, PP -> NI')

        interface.clear_prune_pending_timer()

        interface.set_prune_state(DownstreamState.NoInfo)


    @staticmethod
    def receivedGraft(interface: "TreeInterfaceDownstream", source_ip):
        """
        Receive Graft(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedGraft, PP -> NI')

        interface.clear_prune_pending_timer()

        interface.set_prune_state(DownstreamState.NoInfo)
        interface.send_graft_ack(source_ip)

    @staticmethod
    def PPTexpires(interface: "TreeInterfaceDownstream"):
        """
        PPT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('PPTexpires, PP -> P')
        interface.set_prune_state(DownstreamState.Pruned)
        interface.set_prune_timer(interface.get_received_prune_holdtime() - pim_globals.JP_OVERRIDE_INTERVAL)

        if len(interface.get_interface().neighbors) > 1:
            interface.send_pruneecho()

    @staticmethod
    def PTexpires(interface: "TreeInterfaceDownstream"):
        """
        PT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        #assert False, "PTexpires in state PP"
        return

    @staticmethod
    def is_now_RPF_Interface(interface: "TreeInterfaceDownstream"):
        """
        RPF_Interface(S) becomes I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('is_now_RPF_Interface, PP -> NI')

        interface.clear_prune_pending_timer()

        interface.set_prune_state(DownstreamState.NoInfo)

    @staticmethod
    def send_state_refresh(interface: "TreeInterfaceDownstream"):
        """
        Send State Refresh(S,G) out I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        return

    def __str__(self):
        return "PrunePending"

class Pruned(DownstreamStateABS):
    '''
    Pruned(P)
    The router has received a Prune(S,G) on this interface from a
    downstream neighbor, and the Prune was not overridden. Data from
    S addressed to group G is no longer being forwarded on this
    interface.
    '''

    @staticmethod
    def receivedPrune(interface: "TreeInterfaceDownstream", holdtime):
        """
        Receive Prune(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedPrune, P -> P')
        if holdtime > interface.remaining_prune_timer():
            interface.set_prune_timer(holdtime)

    @staticmethod
    def receivedJoin(interface: "TreeInterfaceDownstream"):
        """
        Receive Join(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedPrune, P -> NI')

        interface.clear_prune_timer()

        interface.set_prune_state(DownstreamState.NoInfo)

    @staticmethod
    def receivedGraft(interface: "TreeInterfaceDownstream", source_ip):
        """
        Receive Graft(S,G)

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('receivedGraft, P -> NI')
        interface.clear_prune_timer()
        interface.set_prune_state(DownstreamState.NoInfo)
        interface.send_graft_ack(source_ip)

    @staticmethod
    def PPTexpires(interface: "TreeInterfaceDownstream"):
        """
        PPT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        #assert False, "PPTexpires in state P"
        return

    @staticmethod
    def PTexpires(interface: "TreeInterfaceDownstream"):
        """
        PT(S,G) Expires

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('PTexpires, P -> NI')
        interface.set_prune_state(DownstreamState.NoInfo)

    @staticmethod
    def is_now_RPF_Interface(interface: "TreeInterfaceDownstream"):
        """
        RPF_Interface(S) becomes I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger('is_now_RPF_Interface, P -> NI')
        interface.clear_prune_timer()
        interface.set_prune_state(DownstreamState.NoInfo)

    @staticmethod
    def send_state_refresh(interface: "TreeInterfaceDownstream"):
        """
        Send State Refresh(S,G) out I

        @type interface: TreeInterfaceDownstreamDownstream
        """
        interface.join_prune_logger.debug('send_state_refresh, P -> P')
        if interface.get_interface().is_state_refresh_capable():
            interface.set_prune_timer(interface.get_received_prune_holdtime())

    def __str__(self):
        return "Pruned"

class DownstreamState():
    NoInfo = NoInfo()
    Pruned = Pruned()
    PrunePending = PrunePending()



import struct

from .PacketPimHello import PacketPimHello
from .PacketPimJoinPrune import PacketPimJoinPrune
from .PacketPimAssert import PacketPimAssert
from .PacketPimGraft import PacketPimGraft
from .PacketPimGraftAck import PacketPimGraftAck
from .PacketPimStateRefresh import PacketPimStateRefresh


from pimdm.utils import checksum
from .PacketPayload import PacketPayload
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|PIM Ver| Type  |   Reserved    |           Checksum            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketPimHeader(PacketPayload):
    PIM_VERSION = 2

    PIM_HDR = "! BB H"
    PIM_HDR_LEN = struct.calcsize(PIM_HDR)

    PIM_MSG_TYPES = {0: PacketPimHello,
                     3: PacketPimJoinPrune,
                     5: PacketPimAssert,
                     6: PacketPimGraft,
                     7: PacketPimGraftAck,
                     9: PacketPimStateRefresh
                     }

    def __init__(self, payload):
        self.payload = payload

    def get_pim_type(self):
        return self.payload.PIM_TYPE

    def bytes(self) -> bytes:
        # obter mensagem e criar checksum
        pim_vrs_type = (PacketPimHeader.PIM_VERSION << 4) + self.get_pim_type()
        msg_without_chcksum = struct.pack(PacketPimHeader.PIM_HDR, pim_vrs_type, 0, 0)
        msg_without_chcksum += self.payload.bytes()
        pim_checksum = checksum(msg_without_chcksum)
        msg = msg_without_chcksum[0:2] + struct.pack("! H", pim_checksum) + msg_without_chcksum[4:]
        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        print("parsePimHdr: ", data)

        pim_hdr = data[0:PacketPimHeader.PIM_HDR_LEN]
        (pim_ver_type, reserved, rcv_checksum) = struct.unpack(PacketPimHeader.PIM_HDR, pim_hdr)

        print(pim_ver_type, reserved, rcv_checksum)
        pim_version = (pim_ver_type & 0xF0) >> 4
        pim_type = pim_ver_type & 0x0F

        if pim_version != PacketPimHeader.PIM_VERSION:
            print("Version of PIM packet received not known (!=2)")
            raise Exception

        msg_to_checksum = data[0:2] + b'\x00\x00' + data[4:]
        if checksum(msg_to_checksum) != rcv_checksum:
            print("wrong checksum")
            print("checksum calculated: " + str(checksum(msg_to_checksum)))
            print("checksum recv: " + str(rcv_checksum))
            raise Exception

        pim_payload = data[PacketPimHeader.PIM_HDR_LEN:]
        pim_payload = PacketPimHeader.PIM_MSG_TYPES[pim_type].parse_bytes(pim_payload)
        return PacketPimHeader(pim_payload)



# IGMP timers (in seconds)
RobustnessVariable = 2
QueryInterval = 125
QueryResponseInterval = 10
MaxResponseTime_QueryResponseInterval = QueryResponseInterval*10
GroupMembershipInterval = RobustnessVariable * QueryInterval + QueryResponseInterval
OtherQuerierPresentInterval = RobustnessVariable * QueryInterval + QueryResponseInterval/2
StartupQueryInterval = QueryInterval / 4
StartupQueryCount = RobustnessVariable
LastMemberQueryInterval = 1
MaxResponseTime_LastMemberQueryInterval = LastMemberQueryInterval*10
LastMemberQueryCount = RobustnessVariable
UnsolicitedReportInterval = 10
Version1RouterPresentTimeout = 400

# IGMP msg type
Membership_Query = 0x11
Version_1_Membership_Report = 0x12
Version_2_Membership_Report = 0x16
Leave_Group = 0x17


import struct
from abc import ABCMeta
import math

class PacketPimHelloOptions(metaclass=ABCMeta):
    PIM_HDR_OPTS = "! HH"
    PIM_HDR_OPTS_LEN = struct.calcsize(PIM_HDR_OPTS)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |              Type             |             Length            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, type: int, length: int):
        self.type = type
        self.length = length

    def bytes(self) -> bytes:
        return struct.pack(PacketPimHelloOptions.PIM_HDR_OPTS, self.type, self.length)

    def __len__(self):
        return self.PIM_HDR_OPTS_LEN + self.length

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        (type, length) = struct.unpack(PacketPimHelloOptions.PIM_HDR_OPTS,
                                        data[:PacketPimHelloOptions.PIM_HDR_OPTS_LEN])
        #print("TYPE:", type)
        #print("LENGTH:", length)
        data = data[PacketPimHelloOptions.PIM_HDR_OPTS_LEN:]
        #return PIM_MSG_TYPES[type](data)
        return PIM_MSG_TYPES.get(type, PacketPimHelloUnknown).parse_bytes(data, type, length)


class PacketPimHelloStateRefreshCapable(PacketPimHelloOptions):
    PIM_HDR_OPT = "! BBH"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |  Version = 1  |   Interval    |            Reserved           |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    VERSION = 1

    def __init__(self, interval: int):
        super().__init__(type=21, length=4)
        self.interval = interval

    def bytes(self) -> bytes:
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.VERSION, self.interval, 0)

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        if type is None or length is None:
            raise Exception
        (version, interval, _) = struct.unpack(PacketPimHelloStateRefreshCapable.PIM_HDR_OPT,
                                                     data[:PacketPimHelloStateRefreshCapable.PIM_HDR_OPT_LEN])
        return PacketPimHelloStateRefreshCapable(interval)



class PacketPimHelloLANPruneDelay(PacketPimHelloOptions):
    PIM_HDR_OPT = "! HH"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |T|       LAN Prune Delay       |       Override Interval       |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, lan_prune_delay: float, override_interval: float):
        super().__init__(type=2, length=4)
        self.lan_prune_delay = 0x7FFF & math.ceil(lan_prune_delay)
        self.override_interval = math.ceil(override_interval)

    def bytes(self) -> bytes:
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.lan_prune_delay, self.override_interval)

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        if type is None or length is None:
            raise Exception
        (lan_prune_delay, override_interval) = struct.unpack(PacketPimHelloLANPruneDelay.PIM_HDR_OPT,
                                                     data[:PacketPimHelloLANPruneDelay.PIM_HDR_OPT_LEN])
        lan_prune_delay = lan_prune_delay & 0x7FFF
        return PacketPimHelloLANPruneDelay(lan_prune_delay=lan_prune_delay, override_interval=override_interval)



class PacketPimHelloHoldtime(PacketPimHelloOptions):
    PIM_HDR_OPT = "! H"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |            Hold Time          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, holdtime: int or float):
        super().__init__(type=1, length=2)
        self.holdtime = int(holdtime)

    def bytes(self) -> bytes:
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.holdtime)

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        if type is None or length is None:
            raise Exception
        (holdtime, ) = struct.unpack(PacketPimHelloHoldtime.PIM_HDR_OPT,
                                     data[:PacketPimHelloHoldtime.PIM_HDR_OPT_LEN])
        #print("HOLDTIME:", holdtime)
        return PacketPimHelloHoldtime(holdtime=holdtime)



class PacketPimHelloGenerationID(PacketPimHelloOptions):
    PIM_HDR_OPT = "! L"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                         Generation ID                         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, generation_id: int):
        super().__init__(type=20, length=4)
        self.generation_id = generation_id

    def bytes(self) -> bytes:
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.generation_id)

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        if type is None or length is None:
            raise Exception
        (generation_id, ) = struct.unpack(PacketPimHelloGenerationID.PIM_HDR_OPT,
                                     data[:PacketPimHelloGenerationID.PIM_HDR_OPT_LEN])
        #print("GenerationID:", generation_id)
        return PacketPimHelloGenerationID(generation_id=generation_id)


class PacketPimHelloUnknown(PacketPimHelloOptions):
    PIM_HDR_OPT = "! L"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                            Unknown                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, type, length):
        super().__init__(type=type, length=length)
        #print("PIM Hello Option Unknown... TYPE=", type, "LENGTH=", length)

    def bytes(self) -> bytes:
        raise Exception

    @staticmethod
    def parse_bytes(data: bytes, type:int = None, length:int = None):
        if type is None or length is None:
            raise Exception
        return PacketPimHelloUnknown(type, length)





PIM_MSG_TYPES = {1: PacketPimHelloHoldtime,
                 2: PacketPimHelloLANPruneDelay,
                 20: PacketPimHelloGenerationID,
                 21: PacketPimHelloStateRefreshCapable,
                 }



import struct
import socket
from .PacketPimEncodedUnicastAddress import PacketPimEncodedUnicastAddress
from .PacketPimEncodedGroupAddress import PacketPimEncodedGroupAddress
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|PIM Ver| Type  |   Reserved    |           Checksum            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Multicast Group Address (Encoded Group Format)        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|             Source Address (Encoded Unicast Format)           |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|           Originator Address (Encoded Unicast Format)         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|R|                     Metric Preference                       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                             Metric                            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|    Masklen    |    TTL        |P|N|O|Reserved |   Interval    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketPimStateRefresh:
    PIM_TYPE = 9

    PIM_HDR_STATE_REFRESH = "! %ss %ss %ss I I BBBB"
    PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES = "! I I BBBB"
    PIM_HDR_STATE_REFRESH_v4 = PIM_HDR_STATE_REFRESH % (PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN)
    PIM_HDR_STATE_REFRESH_v6 = PIM_HDR_STATE_REFRESH % (PacketPimEncodedGroupAddress.PIM_ENCODED_GROUP_ADDRESS_HDR_LEN_IPv6, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN_IPV6, PacketPimEncodedUnicastAddress.PIM_ENCODED_UNICAST_ADDRESS_HDR_LEN_IPV6)

    PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES_LEN = struct.calcsize(PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES)
    PIM_HDR_STATE_REFRESH_v4_LEN = struct.calcsize(PIM_HDR_STATE_REFRESH_v4)
    PIM_HDR_STATE_REFRESH_v6_LEN = struct.calcsize(PIM_HDR_STATE_REFRESH_v6)

    def __init__(self, multicast_group_adress: str or bytes, source_address: str or bytes, originator_adress: str or bytes,
                 metric_preference: int, metric: int, mask_len: int, ttl: int, prune_indicator_flag: bool,
                 prune_now_flag: bool, assert_override_flag: bool, interval: int):

        if type(multicast_group_adress) is bytes:
            multicast_group_adress = socket.inet_ntoa(multicast_group_adress)
        if type(source_address) is bytes:
            source_address = socket.inet_ntoa(source_address)
        if type(originator_adress) is bytes:
            originator_adress = socket.inet_ntoa(originator_adress)

        self.multicast_group_adress = multicast_group_adress
        self.source_address = source_address
        self.originator_adress = originator_adress
        self.metric_preference = metric_preference
        self.metric = metric
        self.mask_len = mask_len
        self.ttl = ttl
        self.prune_indicator_flag = prune_indicator_flag
        self.prune_now_flag = prune_now_flag
        self.assert_override_flag = assert_override_flag
        self.interval = interval

    def bytes(self) -> bytes:
        multicast_group_adress = PacketPimEncodedGroupAddress(self.multicast_group_adress).bytes()
        source_address = PacketPimEncodedUnicastAddress(self.source_address).bytes()
        originator_adress = PacketPimEncodedUnicastAddress(self.originator_adress).bytes()
        prune_and_assert_flags = (self.prune_indicator_flag << 7) | (self.prune_now_flag << 6) | (self.assert_override_flag << 5)

        msg = multicast_group_adress + source_address + originator_adress + \
              struct.pack(self.PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES, 0x7FFFFFFF & self.metric_preference,
                          self.metric, self.mask_len, self.ttl, prune_and_assert_flags, self.interval)

        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        multicast_group_adress_obj = PacketPimEncodedGroupAddress.parse_bytes(data)
        multicast_group_adress_len = len(multicast_group_adress_obj)
        data = data[multicast_group_adress_len:]

        source_address_obj = PacketPimEncodedUnicastAddress.parse_bytes(data)
        source_address_len = len(source_address_obj)
        data = data[source_address_len:]

        originator_address_obj = PacketPimEncodedUnicastAddress.parse_bytes(data)
        originator_address_len = len(originator_address_obj)
        data = data[originator_address_len:]

        (metric_preference, metric, mask_len, ttl, reserved_and_prune_and_assert_flags, interval) = struct.unpack(PacketPimStateRefresh.PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES, data[:PacketPimStateRefresh.PIM_HDR_STATE_REFRESH_WITHOUT_ADDRESSES_LEN])
        metric_preference = 0x7FFFFFFF & metric_preference
        prune_indicator_flag = (0x80 & reserved_and_prune_and_assert_flags) >> 7
        prune_now_flag = (0x40 & reserved_and_prune_and_assert_flags) >> 6
        assert_override_flag = (0x20 & reserved_and_prune_and_assert_flags) >> 5

        pim_payload = PacketPimStateRefresh(multicast_group_adress_obj.group_address, source_address_obj.unicast_address,
                                            originator_address_obj.unicast_address, metric_preference, metric, mask_len,
                                            ttl, prune_indicator_flag, prune_now_flag, assert_override_flag, interval)

        return pim_payload
